import jax
from typing import TYPE_CHECKING
from onnx import helper
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.select_n_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.select_n.html",
    onnx=[
        {
            "component": "Where",
            "doc": "https://onnx.ai/onnx/operators/onnx__Where.html",
        }
    ],
    since="v0.2.0",
    context="primitives.lax",
    testcases=[
        # {
        #     "testcase": "select_n",
        #     "callable": lambda pred, x, y: jax.lax.select_n(pred, x, y),
        #     "input_shapes": [(3,), (3,), (3,)],
        #     "test_data": [(jnp.array([True, False, True]), jnp.array([1, 2, 3]), jnp.array([4, 5, 6]))]
        # },
        # {
        #     "testcase": "select_n_float",
        #     "callable": lambda pred, x, y: jax.lax.select_n(pred, x, y),
        #     "input_shapes": [(2, 2), (2, 2), (2, 2)],
        #     "test_data": [(jnp.array([[True, False], [False, True]]), jnp.array([[1.0, 2.0], [3.0, 4.0]]), jnp.array([[5.0, 6.0], [7.0, 8.0]]))]
        # },
        # {
        #     "testcase": "select_n_int",
        #     "callable": lambda pred, x, y: jax.lax.select_n(pred, x, y),
        #     "input_shapes": [(1, 4), (1, 4), (1, 4)],
        #     "test_data": [(jnp.array([[True, True, False, False]]), jnp.array([[1, 2, 3, 4]]), jnp.array([[5, 6, 7, 8]]))]
        # }
    ],
)
class SelectNPlugin(PrimitivePlugin):
    """Plugin for converting jax.lax.select_n to ONNX Where."""

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX select_n primitive by mapping it to ONNX's Where operator."""
        condition_name = s.get_name(node_inputs[0])
        false_name = s.get_name(node_inputs[1])  # JAX 'x' maps to ONNX 'Y'
        true_name = s.get_name(node_inputs[2])  # JAX 'y' maps to ONNX 'X'
        output_name = s.get_var_name(node_outputs[0])

        node = helper.make_node(
            "Where",
            inputs=[condition_name, true_name, false_name],  # ONNX: condition, X, Y
            outputs=[output_name],
            name=s.get_unique_name("where"),
        )
        s.add_node(node)



import jax
import jax.numpy as jnp
from typing import TYPE_CHECKING
from onnx import helper
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


def gamma(key, alpha):
    d = alpha - 1 / 3
    c = 1 / jnp.sqrt(9 * d)
    z = jax.random.normal(key, alpha.shape)
    v = (1 + c * z) ** 3
    u = jax.random.uniform(key, alpha.shape)
    x = d * v
    acceptance = (v > 0) & (jnp.log(u) < (0.5 * z**2 + d - d * v + d * jnp.log(v)))
    # Re-sample for rejected values
    z = jax.random.normal(key, alpha.shape)
    v = (1 + c * z) ** 3
    x = jnp.where(acceptance, x, d * v)
    # Clip when alpha is zero
    x = jnp.where(alpha == 0, 0.0, x)
    return x


def gamma_log(key, alpha):
    return jnp.log(gamma(key, alpha))


@register_primitive(
    jaxpr_primitive=jax.random.random_gamma_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.random.gamma.html",
    onnx=[],
    since="v0.2.0",
    context="primitives.random",
    testcases=[
        # {
        #     "testcase": "random_gamma_test1",
        #     "callable": lambda alpha: jax.random.gamma(jax.random.PRNGKey(0), alpha),
        #     "input_shapes": [(3,)],
        # }
    ],
)
class RandomGammaPlugin(PrimitivePlugin):
    """Plugin for converting jax.random.random_gamma to ONNX."""

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX random_gamma primitive by constructing a subgraph for gamma sampling."""
        shape = node_inputs[1].aval.shape
        key = jax.random.PRNGKey(0)
        alpha = jnp.zeros(shape)

        subconverter = Jaxpr2OnnxConverter()

        if params.get("log_space", False):
            subconverter.trace_jaxpr(gamma_log, (key, alpha))
        else:
            subconverter.trace_jaxpr(gamma, (key, alpha))

        for outer_var, inner_tensor in zip(node_inputs, subconverter.builder.inputs):
            outer_name = s.get_name(outer_var)
            inner_name = inner_tensor.name
            id_node = helper.make_node(
                "Identity",
                inputs=[outer_name],
                outputs=[inner_name],
                name=s.get_unique_name("gamma_input"),
            )
            s.add_node(id_node)

        s.builder.nodes.extend(subconverter.builder.nodes)
        s.builder.initializers.extend(subconverter.builder.initializers)
        s.builder.name_counter = subconverter.builder.name_counter

        for outer_var, inner_tensor in zip(node_outputs, subconverter.builder.outputs):
            outer_name = s.get_name(outer_var)
            inner_name = inner_tensor.name
            id_node = helper.make_node(
                "Identity",
                inputs=[inner_name],
                outputs=[outer_name],
                name=s.get_unique_name("gamma_output"),
            )
            s.add_node(id_node)



import jax
from typing import TYPE_CHECKING
from onnx import helper
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.reshape_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.reshape.html",
    onnx=[
        {
            "component": "Reshape",
            "doc": "https://onnx.ai/onnx/operators/onnx__Reshape.html",
        }
    ],
    since="v0.2.0",
    context="primitives.lax",
    testcases=[
        {
            "testcase": "reshape",
            "callable": lambda x: jax.lax.reshape(x, (9,)),
            "input_shapes": [(3, 3)],
        }
    ],
)
class ReshapePlugin(PrimitivePlugin):
    """Plugin for converting jax.lax.reshape to ONNX Reshape."""

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX reshape primitive."""
        input_name = s.get_name(node_inputs[0])
        output_name = s.get_var_name(node_outputs[0])
        new_shape = params["new_sizes"]
        input_shape = node_inputs[0].aval.shape

        def _process_newshape(newshape):
            if isinstance(newshape, (int, str)):
                newshape = [newshape]
            else:
                newshape = list(newshape)
            neg_one_count = 0
            for dim in newshape:
                if isinstance(dim, int):
                    if dim == -1:
                        neg_one_count += 1
                    elif dim < 0:
                        raise ValueError("Invalid shape dimension: {}".format(dim))
                elif not isinstance(dim, str):
                    raise ValueError("Invalid shape dimension: {}".format(dim))
            if neg_one_count > 1:
                raise ValueError("Only one dimension can be -1 (inferred).")
            return newshape

        def _concretize_shape(shape, concrete_value=2):
            return tuple(
                concrete_value if isinstance(dim, str) else dim for dim in shape
            )

        processed_newshape = _process_newshape(new_shape)
        concrete_shape = _concretize_shape(processed_newshape)

        if len(new_shape) == 2 and new_shape[0] == 1 and input_shape == (new_shape[1],):
            s.var_to_name[node_outputs[0]] = input_name
            return

        shape_name = s.get_unique_name("reshape_shape")
        s.add_initializer(name=shape_name, vals=concrete_shape)

        node = helper.make_node(
            "Reshape",
            inputs=[input_name, shape_name],
            outputs=[output_name],
            name=s.get_unique_name("reshape"),
        )
        s.add_node(node)



# file: jax2onnx/examples/cnn.py
from flax import nnx
import jax
from jax2onnx.plugin_system import register_example


class CNN(nnx.Module):
    def __init__(self, *, rngs: nnx.Rngs):
        self.conv1 = nnx.Conv(1, 32, kernel_size=(3, 3), rngs=rngs)
        self.conv2 = nnx.Conv(32, 64, kernel_size=(3, 3), rngs=rngs)
        self.avg_pool = lambda x: nnx.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
        self.linear1 = nnx.Linear(3136, 256, rngs=rngs)
        self.linear2 = nnx.Linear(256, 10, rngs=rngs)

    def __call__(self, x: jax.Array):
        x = self.avg_pool(nnx.relu(self.conv1(x)))
        x = self.avg_pool(nnx.relu(self.conv2(x)))
        x = x.reshape(x.shape[0], -1)  # flatten
        x = nnx.relu(self.linear1(x))
        x = self.linear2(x)
        return x


register_example(
    component="CNN",
    description="A simple convolutional neural network (CNN).",
    source="https://github.com/google/flax/blob/main/README.md",
    since="v0.1.0",
    context="examples.nnx",
    children=[
        "nnx.Conv",
        "nnx.Linear",
        "nnx.avg_pool",
        "nnx.relu",
        "lax.reshape",
    ],
    testcases=[
        {
            "testcase": "cnn",
            "callable": CNN(rngs=nnx.Rngs(0)),
            "input_shapes": [(2, 28, 28, 1)],
        }
    ],
)



from jax import core, numpy as jnp
from jax.extend.core import Primitive
from onnx import helper
from typing import TYPE_CHECKING, Tuple, Union, Sequence, Optional
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the Transpose primitive
jnp.transpose_p = Primitive("jnp.transpose")
jnp.transpose_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=jnp.transpose_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.transpose.html",
    onnx=[
        {
            "component": "Transpose",
            "doc": "https://onnx.ai/onnx/operators/onnx__Transpose.html",
        }
    ],
    since="v0.1.0",
    context="primitives.jnp",
    testcases=[
        {
            "testcase": "transpose_basic",
            "callable": lambda a: jnp.transpose(a, axes=(1, 0)),
            "input_shapes": [(2, 3)],
        },
        {
            "testcase": "transpose_reverse",
            "callable": lambda a: jnp.transpose(a, axes=(2, 1, 0)),
            "input_shapes": [(2, 3, 4)],
        },
        {
            "testcase": "transpose_4d",
            "callable": lambda a: jnp.transpose(a, axes=(0, 2, 3, 1)),
            "input_shapes": [(1, 2, 3, 4)],
        },
        {
            "testcase": "transpose_square_matrix",
            "callable": lambda a: jnp.transpose(a, axes=(1, 0)),
            "input_shapes": [(5, 5)],
        },
        {
            "testcase": "transpose_high_dim",
            "callable": lambda a: jnp.transpose(a, axes=(4, 3, 2, 1, 0)),
            "input_shapes": [(2, 3, 4, 5, 6)],
        },
        {
            "testcase": "transpose_no_axes",  # Test case for default axes (reversal)
            "callable": lambda a: jnp.transpose(a),
            "input_shapes": [(2, 3, 4)],
        },
        {
            "testcase": "transpose_dynamic",
            "callable": lambda a: jnp.transpose(a, axes=(0, 2, 1)),
            "input_shapes": [("B", 3, 4)],  # Dynamic batch dimension
        },
    ],
)
class TransposePlugin(PrimitivePlugin):
    """
    Plugin for converting jax.numpy.transpose to ONNX.
    """

    @staticmethod
    def _transpose_abstract_eval(x, axes: Optional[Tuple[int, ...]]):
        """Computes the output shape for jnp.transpose."""
        x_shape = list(x.shape)
        if axes is None:
            axes = tuple(reversed(range(len(x_shape))))
        if len(axes) != len(x_shape):
            raise ValueError(
                f"Axes length {len(axes)} does not match input rank {len(x_shape)}"
            )

        output_shape = [x_shape[i] for i in axes]
        return core.ShapedArray(tuple(output_shape), x.dtype)

    @staticmethod
    def abstract_eval(x, axes: Optional[Tuple[int, ...]]):
        return TransposePlugin._transpose_abstract_eval(x, axes)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles ONNX conversion for jnp.transpose."""
        axes = params["axes"]
        input_name = s.get_name(node_inputs[0])
        output_name = s.get_name(node_outputs[0])
        input_shape = node_inputs[0].aval.shape

        # If axes is None, default to reversing the axes.
        if axes is None:
            axes = tuple(reversed(range(len(input_shape))))
        # This part is not needed, the bind function already manage that
        # elif isinstance(axes, int):
        #     n = len(input_shape)
        #     axes = (axes,) + tuple(i for i in range(n) if i != axes)
        else:
            axes = tuple(axes)  # Ensure axes is a tuple.

        transpose_node = helper.make_node(
            "Transpose",
            inputs=[input_name],
            outputs=[output_name],
            perm=list(axes),  # ONNX expects a list
            name=s.get_unique_name("transpose"),
        )
        s.add_node(transpose_node)

        output_shape = tuple(input_shape[i] for i in axes)
        s.add_shape_info(output_name, output_shape)

    @staticmethod
    def _transpose(a, axes: Optional[Union[Sequence[int], int]] = None):
        """Defines the primitive binding for Transpose."""
        n = len(a.shape)
        if axes is None:
            axes = tuple(reversed(range(n)))
        elif isinstance(axes, int):
            axes = (axes,) + tuple(i for i in range(n) if i != axes)
        else:
            axes = tuple(axes)
        if len(axes) != n:
            raise ValueError(f"Axes length {len(axes)} does not match input rank {n}")
        return jnp.transpose_p.bind(a, axes=axes)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for Transpose."""

        def patched_transpose(a, axes: Optional[Union[Sequence[int], int]] = None):
            return TransposePlugin._transpose(a, axes)

        return patched_transpose

    @staticmethod
    def patch_info():
        """Provides patching information for Transpose."""
        return {
            "patch_targets": [jnp],
            "patch_function": lambda _: TransposePlugin.get_monkey_patch(),
            "target_attribute": "transpose",
        }


# Register abstract evaluation function
jnp.transpose_p.def_abstract_eval(TransposePlugin.abstract_eval)



import numpy as np
import jax
from typing import TYPE_CHECKING
from onnx import helper
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.reduce_min_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.reduce_min.html",
    onnx=[
        {
            "component": "ReduceMin",
            "doc": "https://onnx.ai/onnx/operators/onnx__ReduceMin.html",
        }
    ],
    since="v0.2.0",
    context="primitives.lax",
    testcases=[
        {
            "testcase": "reduce_min",
            "callable": lambda x: jax.lax.reduce_min(x, axes=(0,)),
            "input_shapes": [(3, 3)],
        }
    ],
)
class ReduceMinPlugin(PrimitivePlugin):
    """Plugin for converting jax.lax.reduce_min to ONNX ReduceMin."""

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX reduce_min primitive."""
        input_name = s.get_name(node_inputs[0])
        output_name = s.get_var_name(node_outputs[0])
        axes = params["axes"]
        axes_name = s.get_constant_name(np.array(axes, dtype=np.int64))
        node = helper.make_node(
            "ReduceMin",
            inputs=[input_name, axes_name],
            outputs=[output_name],
            name=s.get_unique_name("reduce_min"),
            keepdims=0 if not params.get("keepdims", False) else 1,
        )
        s.add_node(node)



from jax import core
from jax.extend.core import Primitive
from flax import nnx
from onnx import helper
from typing import TYPE_CHECKING
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin
import numpy as np

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the LayerNorm primitive
nnx.layer_norm_p = Primitive("nnx.layer_norm")
nnx.layer_norm_p.multiple_results = False  # Correctly set at initialization


@register_primitive(
    jaxpr_primitive=nnx.layer_norm_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/normalization.html#flax.nnx.LayerNorm",
    onnx=[
        {
            "component": "LayerNormalization",
            "doc": "https://onnx.ai/onnx/operators/onnx__LayerNormalization.html",
        }
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "layer_norm",
            "callable": nnx.LayerNorm(num_features=32, epsilon=1e-5, rngs=nnx.Rngs(0)),
            "input_shapes": [(10, 20, 32)],
        },
        {
            "testcase": "layer_norm_multiaxis",
            "callable": nnx.LayerNorm(
                3 * 3 * 64,
                reduction_axes=(1, 2, 3),
                feature_axes=(1, 2, 3),
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(1, 3, 3, 64)],
        },
    ],
)
class LayerNormPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.LayerNorm to ONNX.
    """

    @staticmethod
    def abstract_eval(x, scale, bias, epsilon, axis):
        """Abstract evaluation function for LayerNorm."""
        return core.ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of LayerNorm to ONNX format."""
        # Expect node_inputs: [x, scale, bias]
        input_name = s.get_name(node_inputs[0])
        scale_name = s.get_name(node_inputs[1]) if node_inputs[1] is not None else None
        bias_name = s.get_name(node_inputs[2]) if node_inputs[2] is not None else None
        output_name = s.get_name(node_outputs[0])

        epsilon = params.get("epsilon")
        axis = params.get("axis", -1)  # Default normalization axis: last dimension

        # ONNX LayerNormalization expects three inputs: input, scale, bias.
        # Handle optional scale and bias.
        inputs = [input_name]
        if scale_name is not None:
            inputs.append(scale_name)
        else:  # If scale is None, a 1s tensor with proper shape must be created.
            input_shape = node_inputs[0].aval.shape
            scale_shape = [1] * len(input_shape)
            scale_shape[axis] = input_shape[axis]
            scale_name = s.builder.get_constant_name(
                np.ones(scale_shape, dtype=np.float32)
            )
            inputs.append(scale_name)

        if bias_name is not None:
            inputs.append(bias_name)
        else:  # If bias is None, a 0s tensor with proper shape must be created.
            input_shape = node_inputs[0].aval.shape
            bias_shape = [1] * len(input_shape)
            bias_shape[axis] = input_shape[axis]
            bias_name = s.builder.get_constant_name(
                np.zeros(bias_shape, dtype=np.float32)
            )
            inputs.append(bias_name)

        ln_node = helper.make_node(
            "LayerNormalization",
            inputs=inputs,
            outputs=[output_name],
            name=s.get_unique_name("layer_norm"),
            axis=axis,
            epsilon=epsilon,
        )
        s.add_node(ln_node)

    @staticmethod
    def _layer_norm(x, scale, bias, epsilon, axis):
        """Defines the primitive binding for LayerNorm."""
        return nnx.layer_norm_p.bind(
            x,
            scale,
            bias,
            epsilon=epsilon,
            axis=axis,
        )

    @staticmethod
    def get_monkey_patch():
        """Returns a patched version of LayerNorm's call method."""

        def patched_layer_norm_call(self, x):
            # Default to axis=-1 if no reduction_axes are provided.
            norm_axis = -1
            if hasattr(self, "reduction_axes"):
                # If reduction_axes is iterable (list/tuple), take the minimum; otherwise, use it directly.
                if isinstance(self.reduction_axes, (list, tuple)):
                    norm_axis = min(self.reduction_axes)
                else:
                    norm_axis = self.reduction_axes
            return LayerNormPlugin._layer_norm(
                x,
                self.scale.value if self.scale is not None else None,
                self.bias.value if self.bias is not None else None,
                epsilon=self.epsilon,
                axis=norm_axis,
            )

        return patched_layer_norm_call

    @staticmethod
    def patch_info():
        """Provides patching information for LayerNorm."""
        return {
            "patch_targets": [nnx.LayerNorm],
            "patch_function": lambda _: LayerNormPlugin.get_monkey_patch(),
            "target_attribute": "__call__",
        }


# Register abstract evaluation function
nnx.layer_norm_p.def_abstract_eval(LayerNormPlugin.abstract_eval)



import jax
import numpy as np
from typing import TYPE_CHECKING
from onnx import helper
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.dot_general_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.dot_general.html",
    onnx=[
        {
            "component": "MatMul",  # Corrected: MatMul is used, not Gemm directly
            "doc": "https://onnx.ai/onnx/operators/onnx__MatMul.html",
        }
    ],
    since="v0.2.0",
    context="primitives.lax",
    testcases=[
        {
            "testcase": "dot_general",
            "callable": lambda x1, x2: jax.lax.dot_general(
                x1, x2, (((1,), (0,)), ((), ()))
            ),
            "input_shapes": [(3, 3), (3, 3)],
        }
    ],
)
class DotGeneralPlugin(PrimitivePlugin):
    """
    Plugin for converting jax.lax.dot_general to ONNX.
    """

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX dot_general primitive with a reshape-Gemm-reshape pattern."""
        input_names = [s.get_name(inp) for inp in node_inputs]
        output_name = s.get_var_name(node_outputs[0])

        # Extract dot_general parameters
        dimension_numbers = params["dimension_numbers"]
        ((lhs_contract, rhs_contract), (lhs_batch, rhs_batch)) = dimension_numbers

        lhs_name, rhs_name = input_names
        lhs_shape = node_inputs[0].aval.shape
        rhs_shape = node_inputs[1].aval.shape
        output_shape = node_outputs[0].aval.shape

        # Compute batch and feature dimensions
        np.prod(lhs_shape[: len(lhs_shape) - len(lhs_contract)], dtype=np.int64)
        feature_size = np.prod(
            lhs_shape[len(lhs_shape) - len(lhs_contract) :], dtype=np.int64
        )
        rhs_output_size = np.prod(
            rhs_shape[len(rhs_shape) - len(rhs_contract) :], dtype=np.int64
        )

        lhs_reshape_name = s.get_unique_name("reshape_input")
        const_shape = np.array([feature_size, rhs_output_size], dtype=np.int64)
        const_name = s.get_constant_name(const_shape)
        node_lhs = helper.make_node(
            "Reshape",
            inputs=[lhs_name, const_name],
            outputs=[lhs_reshape_name],
            name=s.get_unique_name("reshape_lhs"),
        )
        s.add_node(node_lhs)

        rhs_reshape_name = s.get_unique_name("rhs_reshape")
        const_name_rhs = s.get_constant_name(
            np.array([feature_size, rhs_output_size], dtype=np.int64)
        )
        node_rhs = helper.make_node(
            "Reshape",
            inputs=[rhs_name, const_name_rhs],
            outputs=[rhs_reshape_name],
            name=s.get_unique_name("reshape_rhs"),
        )
        s.add_node(node_rhs)

        gemm_output_name = s.get_unique_name("gemm_output")
        gemm_node = helper.make_node(
            "Gemm",
            inputs=[lhs_reshape_name, rhs_reshape_name],
            outputs=[gemm_output_name],
            name=s.get_unique_name("gemm"),
            alpha=1.0,
            beta=1.0,
            transB=0,
        )
        s.add_node(gemm_node)

        reshape_output_node = helper.make_node(
            "Reshape",
            inputs=[
                gemm_output_name,
                s.get_constant_name(np.array(output_shape, dtype=np.int64)),
            ],
            outputs=[output_name],
            name=s.get_unique_name("reshape_output"),
        )
        s.add_node(reshape_output_node)



import jax
import numpy as np
from typing import TYPE_CHECKING
from onnx import helper
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.broadcast_in_dim_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.broadcast_in_dim.html",
    onnx=[
        {
            "component": "Expand",
            "doc": "https://onnx.ai/onnx/operators/onnx__Expand.html",
        }
    ],
    since="v0.2.0",
    context="primitives.lax",
    testcases=[
        {
            "testcase": "broadcast_in_dim",
            "callable": lambda x: jax.lax.broadcast_in_dim(
                x, (3,), broadcast_dimensions=(0,)
            ),
            "input_shapes": [(3,)],
        },
        {
            "testcase": "broadcast_in_dim_2d_to_3d",
            "callable": lambda x: jax.lax.broadcast_in_dim(
                x, (2, 3, 4), broadcast_dimensions=(1, 2)
            ),
            "input_shapes": [(3, 4)],
        },
        {
            "testcase": "broadcast_in_dim_scalar",
            "callable": lambda x: jax.lax.broadcast_in_dim(
                x, (2, 3, 4), broadcast_dimensions=()
            ),
            "input_shapes": [()],
        },
    ],
)
class BroadcastInDimPlugin(PrimitivePlugin):
    """
    Plugin for converting jax.lax.broadcast_in_dim to ONNX.
    """

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX broadcast_in_dim primitive."""
        input_name = s.get_name(node_inputs[0])
        output_name = s.get_var_name(node_outputs[0])
        broadcast_dimensions = params["broadcast_dimensions"]
        shape = params["shape"]
        shape_name = s.get_constant_name(np.array(shape, dtype=np.int64))
        # First, reshape input to add singleton dimensions where necessary.
        reshape_output = s.get_unique_name("reshape_output")
        reshape_shape = []
        idx = 0
        for i in range(len(shape)):
            if i in broadcast_dimensions:
                if idx < len(node_inputs[0].aval.shape):
                    reshape_shape.append(node_inputs[0].aval.shape[idx])
                else:
                    reshape_shape.append(1)
                idx += 1
            else:
                reshape_shape.append(1)
        reshape_shape_name = s.get_constant_name(
            np.array(reshape_shape, dtype=np.int64)
        )
        node_reshape = helper.make_node(
            "Reshape",
            inputs=[input_name, reshape_shape_name],
            outputs=[reshape_output],
            name=s.get_unique_name("reshape_for_broadcast"),
        )
        s.add_node(node_reshape)
        # Then, expand to the target shape.
        node_expand = helper.make_node(
            "Expand",
            inputs=[reshape_output, shape_name],
            outputs=[output_name],
            name=s.get_unique_name("expand"),
        )
        s.add_node(node_expand)



from jax import core
from jax.extend.core import Primitive
from flax import nnx
from onnx import helper
from typing import TYPE_CHECKING
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the Sigmoid primitive
nnx.sigmoid_p = Primitive("nnx.sigmoid")
nnx.sigmoid_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=nnx.sigmoid_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/activations.html#flax.nnx.sigmoid",
    onnx=[
        {
            "component": "Sigmoid",
            "doc": "https://onnx.ai/onnx/operators/onnx__Sigmoid.html",
        }
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "sigmoid",
            "callable": lambda x: nnx.sigmoid(x),
            "input_shapes": [(3,)],
        }
    ],
)
class SigmoidPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.sigmoid to ONNX.
    """

    @staticmethod
    def abstract_eval(x):
        """Abstract evaluation function for Sigmoid."""
        return core.ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of Sigmoid to ONNX format."""
        input_var = node_inputs[0]
        output_var = node_outputs[0]

        input_name = s.get_name(input_var)
        output_name = s.get_name(output_var)

        sigmoid_node = helper.make_node(
            "Sigmoid",
            inputs=[input_name],
            outputs=[output_name],
            name=s.get_unique_name("sigmoid"),
        )
        s.add_node(sigmoid_node)

    @staticmethod
    def _sigmoid(x):
        """Defines the primitive binding for Sigmoid."""
        return nnx.sigmoid_p.bind(x)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for Sigmoid."""

        def patched_sigmoid(x):
            return SigmoidPlugin._sigmoid(x)

        return patched_sigmoid

    @staticmethod
    def patch_info():
        """Provides patching information for Sigmoid."""
        return {
            "patch_targets": [nnx],
            "patch_function": lambda _: SigmoidPlugin.get_monkey_patch(),
            "target_attribute": "sigmoid",
        }


# Register abstract evaluation function
nnx.sigmoid_p.def_abstract_eval(SigmoidPlugin.abstract_eval)



# file: jax2onnx/plugins/flax/nnx/linear_general.py
"""
Linear General Plugin for JAX to ONNX conversion.

This plugin enables conversion of flax.nnx.LinearGeneral layers to ONNX format.
It transforms JAX’s linear_general operations (a specialized dot_general for linear layers)
into an ONNX Gemm operator with necessary Reshape operations.

The conversion process involves:
  1. Calculating the output shape and the reshaping parameters.
  2. Providing an abstract evaluation for JAX's tracing system.
  3. Converting the operation to ONNX using Gemm and Reshape nodes.
  4. Monkey-patching LinearGeneral.__call__ to redirect calls to our primitive.
"""

import numpy as np
from jax import core
from flax import nnx
from onnx import helper
from jax.extend.core import Primitive
from typing import TYPE_CHECKING
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the primitive for linear_general operations.
nnx.linear_general_p = Primitive("nnx.linear_general")


@register_primitive(
    jaxpr_primitive=nnx.linear_general_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/linear.html#flax.nnx.LinearGeneral",
    onnx=[
        {"component": "Gemm", "doc": "https://onnx.ai/onnx/operators/onnx__Gemm.html"},
        {
            "component": "Reshape",
            "doc": "https://onnx.ai/onnx/operators/onnx__Reshape.html",
        },
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "linear_general",
            "callable": nnx.LinearGeneral(
                in_features=(8, 32),
                out_features=(256,),
                axis=(-2, -1),
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [("B", 4, 8, 32)],
        },
        {
            "testcase": "linear_general_2",
            "callable": nnx.LinearGeneral(
                in_features=(30,),
                out_features=(20,),
                axis=(-1,),
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(3, 30)],
        },
        {
            "testcase": "linear_general_3",
            "callable": nnx.LinearGeneral(
                in_features=(256,),
                out_features=(8, 32),
                axis=(-1,),
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(2, 4, 256)],
        },
        {
            "testcase": "linear_general_4",
            "callable": nnx.LinearGeneral(
                in_features=(8, 32),
                out_features=(256,),
                axis=(-2, -1),
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(2, 4, 8, 32)],
        },
    ],
)
class LinearGeneralPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.LinearGeneral to ONNX.

    Converts a LinearGeneral operation into a Gemm (matrix multiplication)
    followed by a Reshape to recover the desired output shape.
    """

    @staticmethod
    def _normalize_contracting_dims(dimension_numbers, x_shape, kernel_shape):
        # Unpack and normalize contracting dimensions to positive indices.
        ((lhs_contract, rhs_contract), _) = dimension_numbers
        lhs_contract = [d % len(x_shape) for d in lhs_contract]
        rhs_contract = [d % len(kernel_shape) for d in rhs_contract]
        return lhs_contract, rhs_contract

    @staticmethod
    def _compute_batch_and_kernel_output_dims(
        x_shape, kernel_shape, lhs_contract, rhs_contract
    ):
        # Compute sizes for batch dimensions from input and non-contracted (output) dimensions from kernel.
        x_batch_dims = [i for i in range(len(x_shape)) if i not in lhs_contract]
        x_batch_dims_sizes = [x_shape[i] for i in x_batch_dims]
        kernel_noncontract_dims = [
            i for i in range(len(kernel_shape)) if i not in rhs_contract
        ]
        kernel_out_dims = [kernel_shape[i] for i in kernel_noncontract_dims]
        return x_batch_dims_sizes, kernel_out_dims

    @staticmethod
    def _shape_linear_general(x_shape, kernel_shape, dimension_numbers):
        """Calculate all reshaping parameters for the Gemm transformation."""
        lhs_contract, rhs_contract = LinearGeneralPlugin._normalize_contracting_dims(
            dimension_numbers, x_shape, kernel_shape
        )
        x_batch_dims_sizes, kernel_out_dims = (
            LinearGeneralPlugin._compute_batch_and_kernel_output_dims(
                x_shape, kernel_shape, lhs_contract, rhs_contract
            )
        )

        output_shape = tuple(x_batch_dims_sizes + kernel_out_dims)
        new_kernel_dims_sizes = (
            np.prod([kernel_shape[i] for i in rhs_contract]).item(),
            np.prod(kernel_out_dims).item(),
        )
        input_gemm_shape = (
            np.prod(x_batch_dims_sizes).item(),
            np.prod([x_shape[i] for i in lhs_contract]).item(),
        )
        output_gemm_shape = (input_gemm_shape[0], new_kernel_dims_sizes[1])

        return {
            "input": x_shape,
            "input_gemm": input_gemm_shape,
            "output_gemm": output_gemm_shape,
            "output": output_shape,
            "new_kernel": new_kernel_dims_sizes,
        }

    @staticmethod
    def _is_noop_reshape(original_shape, target_shape):
        # A reshape is a no-op if the shapes are identical except possibly the batch dimension.
        return (
            len(original_shape) == len(target_shape)
            and original_shape[1:] == target_shape[1:]
        )

    @staticmethod
    def abstract_eval(x, kernel, bias, dimension_numbers):
        """Abstract evaluation: computes output shape and dtype."""
        shapes = LinearGeneralPlugin._shape_linear_general(
            x.shape, kernel.shape, dimension_numbers
        )
        return core.ShapedArray(shapes["output"], x.dtype)

    def to_onnx(
        self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, dimension_params
    ):
        """Convert linear_general operation to ONNX format."""
        input_var, kernel_var, bias_var = node_inputs[:3]
        output_var = node_outputs[0]

        input_name = s.get_name(input_var)
        output_name = s.get_name(output_var)
        kernel_name = s.get_name(kernel_var)
        bias_name = s.get_name(bias_var) if bias_var else None

        shape_info = LinearGeneralPlugin._shape_linear_general(
            input_var.aval.shape,
            kernel_var.aval.shape,
            dimension_params["dimension_numbers"],
        )
        output_shape = shape_info["output"]
        new_kernel_shape = shape_info["new_kernel"]
        input_gemm_shape = shape_info["input_gemm"]
        output_gemm_shape = shape_info["output_gemm"]

        kernel_const = s.name_to_const[kernel_name]
        weights_name = s.get_constant_name(kernel_const.reshape(new_kernel_shape))

        target_input_shape = (-1,) + input_gemm_shape[1:]
        if LinearGeneralPlugin._is_noop_reshape(
            input_var.aval.shape, target_input_shape
        ):
            input_reshape_name = input_name
        else:
            input_reshape_name = s.get_unique_name("input_reshape")
            s.add_node(
                helper.make_node(
                    "Reshape",
                    inputs=[
                        input_name,
                        s.get_constant_name(
                            np.array(target_input_shape, dtype=np.int64)
                        ),
                    ],
                    outputs=[input_reshape_name],
                    name=s.get_unique_name("reshape_input"),
                )
            )
            s.add_shape_info(input_reshape_name, input_gemm_shape)

        # Prepare bias: reshape if necessary or create zero bias.
        if bias_name is not None:
            bias_const = s.name_to_const[bias_name]
            target_bias_shape = (output_gemm_shape[1],)
            if bias_const.shape != target_bias_shape:
                bias_const = bias_const.reshape(target_bias_shape)
                bias_name = s.get_constant_name(bias_const)
            gemm_inputs = [input_reshape_name, weights_name, bias_name]
        else:
            bias_shape = (output_gemm_shape[1],)
            zero_bias = np.zeros(bias_shape, dtype=input_var.aval.dtype)
            bias_name = s.get_constant_name(zero_bias)
            gemm_inputs = [input_reshape_name, weights_name, bias_name]

        gemm_output_name = (
            output_name
            if LinearGeneralPlugin._is_noop_reshape(output_gemm_shape, output_shape)
            else s.get_unique_name("gemm_output")
        )
        s.add_node(
            helper.make_node(
                "Gemm",
                inputs=gemm_inputs,
                outputs=[gemm_output_name],
                name=s.get_unique_name("gemm"),
            )
        )
        s.add_shape_info(gemm_output_name, output_gemm_shape)

        if gemm_output_name != output_name:
            target_output_shape = [-1] + list(output_shape[1:])
            s.add_node(
                helper.make_node(
                    "Reshape",
                    inputs=[
                        gemm_output_name,
                        s.get_constant_name(
                            np.array(target_output_shape, dtype=np.int64)
                        ),
                    ],
                    outputs=[output_name],
                    name=s.get_unique_name("reshape_output"),
                )
            )

    @staticmethod
    def _linear_general(x, kernel, bias, dimension_numbers):
        nnx.linear_general_p.multiple_results = False
        return nnx.linear_general_p.bind(
            x, kernel, bias, dimension_numbers=dimension_numbers
        )

    @staticmethod
    def linear_general(x, kernel, bias, dimension_numbers):
        """Binding function for linear_general."""
        return LinearGeneralPlugin._linear_general(x, kernel, bias, dimension_numbers)

    @staticmethod
    def get_monkey_patch():
        """Returns a patched version of LinearGeneral.__call__."""

        def patched_linear_general_call(self, x):
            contracting_dims = (
                (self.axis,) if isinstance(self.axis, int) else self.axis,
                tuple(range(len(self.in_features))),
            )
            dimension_numbers = (contracting_dims, ((), ()))
            return LinearGeneralPlugin._linear_general(
                x,
                self.kernel.value,
                self.bias.value if self.bias else None,
                dimension_numbers,
            )

        return patched_linear_general_call

    @staticmethod
    def patch_info():
        """Provides patching information."""
        return {
            "patch_targets": [nnx.LinearGeneral],
            "patch_function": lambda _: LinearGeneralPlugin.get_monkey_patch(),
        }


# Register abstract evaluation function.
nnx.linear_general_p.def_abstract_eval(LinearGeneralPlugin.abstract_eval)



from jax2onnx import converter  # noqa: F401
from jax2onnx import tester  # noqa: F401
from jax2onnx.converter.converter import save_onnx  # noqa: F401
from jax2onnx.tester import allclose  # noqa: F401



import jax
import numpy as np
from typing import TYPE_CHECKING
from onnx import helper, TensorProto
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.dynamic_slice_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.dynamic_slice.html",
    onnx=[
        {
            "component": "Slice",
            "doc": "https://onnx.ai/onnx/operators/onnx__Slice.html",
        }
    ],
    since="v0.1.0",
    context="primitives.lax",
    testcases=[
        {
            "testcase": "dynamic_slice_test1",
            "callable": lambda x: jax.lax.dynamic_slice(x, [1], [2]),
            "input_shapes": [(5,)],
        },
        {  # Added 2D test case
            "testcase": "dynamic_slice_2d",
            "callable": lambda x: jax.lax.dynamic_slice(x, (1, 2), (2, 3)),
            "input_shapes": [(4, 6)],
        },
        {  # Added 3D test case
            "testcase": "dynamic_slice_3d",
            "callable": lambda x: jax.lax.dynamic_slice(x, (1, 0, 2), (2, 3, 1)),
            "input_shapes": [(3, 4, 5)],
        },
    ],
)
class DynamicSlicePlugin(PrimitivePlugin):
    """Plugin for converting jax.lax.dynamic_slice to ONNX Slice."""

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX dynamic slice primitive.

        In JAX, the dynamic slice primitive is bound as:
          dynamic_slice_p.bind(operand, *start_indices, *dynamic_sizes, slice_sizes=static_sizes)
        Here we assume that no dynamic_sizes are provided so that
          node_inputs = [operand, start_index0, start_index1, ..., start_index{d-1}],
        where d is the rank of the operand.

        ONNX Slice requires a 1-D 'starts' input. We therefore:
          1. Process each start index by unsqueezing it via an Unsqueeze node,
             then casting it to int64.
          2. Concatenate these casted start indices to form a single 1-D tensor.
          3. Create a constant for the static slice sizes (as int64).
          4. Compute ends = starts + slice_sizes via an Add node.
          5. Create an axes constant (as int64) for all dimensions and build the Slice node.
        """
        # Get the operand to slice.
        operand_name = s.get_name(node_inputs[0])
        output_name = s.get_var_name(node_outputs[0])

        # Determine the rank (number of dimensions) from the operand's shape.
        d = len(node_inputs[0].aval.shape)

        # Process each dynamic start index (from node_inputs[1:1+d]).
        start_names = []
        for i in range(1, 1 + d):
            start_i_name = s.get_name(node_inputs[i])
            # Create a constant for axes: [0] with int64 type.
            axes_const = s.get_constant_name(np.array([0], dtype=np.int64))
            unsqueezed_name = s.get_unique_name(f"unsqueezed_start_{i}")
            unsqueeze_node = helper.make_node(
                "Unsqueeze",
                inputs=[start_i_name, axes_const],
                outputs=[unsqueezed_name],
                name=s.get_unique_name("unsqueeze_start"),
            )
            s.add_node(unsqueeze_node)

            # Insert a Cast node to ensure type int64.
            casted_name = s.get_unique_name(f"cast_start_{i}")
            cast_node = helper.make_node(
                "Cast",
                inputs=[unsqueezed_name],
                outputs=[casted_name],
                name=s.get_unique_name("cast_start"),
                to=TensorProto.INT64,
            )
            s.add_node(cast_node)

            start_names.append(casted_name)

        # Concatenate the casted start indices into one 1-D tensor.
        starts_concat_name = s.get_unique_name("dynamic_starts")
        concat_node = helper.make_node(
            "Concat",
            inputs=start_names,
            outputs=[starts_concat_name],
            name=s.get_unique_name("concat_starts"),
            axis=0,
        )
        s.add_node(concat_node)

        # Create constant for static slice sizes as int64.
        slice_sizes = params["slice_sizes"]
        slice_sizes_const = s.get_constant_name(np.array(slice_sizes, dtype=np.int64))

        # Compute dynamic ends: ends = starts + slice_sizes.
        ends_name = s.get_unique_name("dynamic_ends")
        add_node = helper.make_node(
            "Add",
            inputs=[starts_concat_name, slice_sizes_const],
            outputs=[ends_name],
            name=s.get_unique_name("add_slice_ends"),
        )
        s.add_node(add_node)

        # Create constant for axes: [0, 1, ..., d-1] as int64.
        axes = list(range(d))
        axes_const = s.get_constant_name(np.array(axes, dtype=np.int64))

        inputs_list = [operand_name, starts_concat_name, ends_name, axes_const]
        if "strides" in params and params["strides"]:
            strides = params["strides"]
            strides_const = s.get_constant_name(np.array(strides, dtype=np.int64))
            inputs_list.append(strides_const)

        node = helper.make_node(
            "Slice",
            inputs=inputs_list,
            outputs=[output_name],
            name=s.get_unique_name("dynamic_slice"),
        )
        s.add_node(node)



import numpy as np
from jax import core, numpy as jnp
from jax.extend.core import Primitive
from onnx import helper
from typing import TYPE_CHECKING, Tuple, List, Union, Dict
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the Einsum primitive
jnp.einsum_p = Primitive("jnp.einsum")
jnp.einsum_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=jnp.einsum_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.einsum.html",
    onnx=[
        {
            "component": "Einsum",
            "doc": "https://onnx.ai/onnx/operators/onnx__Einsum.html",
        }
    ],
    since="v0.1.0",
    context="primitives.jnp",
    testcases=[
        {
            "testcase": "einsum",
            "callable": lambda a, b: jnp.einsum("ij,j->i", a, b, precision=None),
            "input_shapes": [(3, 3), (3,)],
        },
        {
            "testcase": "einsum_matmul",
            "callable": lambda a, b: jnp.einsum("ij,jk->ik", a, b, precision=None),
            "input_shapes": [(4, 3), (3, 5)],
        },
        {
            "testcase": "einsum_dynamic",
            "callable": lambda a, b: jnp.einsum("ij,j->i", a, b, precision=None),
            "input_shapes": [("B", 3), (3,)],
        },
        {
            "testcase": "einsum_dynamic_matmul",
            "callable": lambda a, b: jnp.einsum("bij,jk->bik", a, b, precision=None),
            "input_shapes": [("B", 5, 3), (3, 4)],
        },
        {
            "testcase": "einsum_transpose",
            "callable": lambda a: jnp.einsum("ij->ji", a, precision=None),
            "input_shapes": [(2, 3)],
        },
        {
            "testcase": "einsum_dynamic_transpose",
            "callable": lambda a: jnp.einsum("bij->bji", a, precision=None),
            "input_shapes": [("B", 2, 3)],
        },
        {
            "testcase": "einsum_dynamic_matmul2",
            "callable": lambda a, b: jnp.einsum("bij,jk->bik", a, b, precision=None),
            "input_shapes": [("B", 5, 3), (3, 4)],
        },
        {
            "testcase": "einsum_dynamic_matmul3",
            "callable": lambda a, b: jnp.einsum("bij,bjk->bik", a, b, precision=None),
            "input_shapes": [("B", 5, 3), ("B", 3, 4)],
        },
    ],
)
class EinsumPlugin(PrimitivePlugin):
    """
    Plugin for converting jax.numpy.einsum to ONNX.
    """

    @staticmethod
    def _parse_einsum_equation(equation: str) -> Tuple[List[str], str]:
        """Parses the einsum equation into input and output terms."""
        parts = equation.split("->")
        if len(parts) != 2:
            raise ValueError("Einsum equation must contain '->'.")
        input_terms, output_term = parts
        return input_terms.split(","), output_term

    @staticmethod
    def _get_dynamic_output_shape(
        input_shapes: List[Tuple[Union[int, str], ...]], equation: str
    ) -> Tuple[Union[int, str], ...]:
        """Calculates the output shape while handling dynamic dimensions."""

        dummy_inputs = [
            np.zeros([1 if isinstance(d, str) else d for d in shape])
            for shape in input_shapes
        ]
        dummy_output = np.einsum(equation, *dummy_inputs)
        output_shape = list(dummy_output.shape)

        input_terms, output_term = EinsumPlugin._parse_einsum_equation(equation)
        index_to_label: Dict[str, Union[int, str]] = {}

        for term, shape in zip(input_terms, input_shapes):
            for i, label in enumerate(term):
                if label not in index_to_label:
                    try:
                        index_to_label[label] = shape[i]
                    except IndexError:
                        index_to_label[label] = -1

        for i, label in enumerate(output_term):
            if label in index_to_label and isinstance(index_to_label[label], str):
                output_shape[i] = index_to_label[label]

        return tuple(output_shape)

    @staticmethod
    def abstract_eval(*operands, equation, precision):
        """Abstract evaluation function for Einsum."""
        input_shapes = [op.shape for op in operands]
        output_shape = EinsumPlugin._get_dynamic_output_shape(input_shapes, equation)
        return core.ShapedArray(output_shape, operands[0].dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of Einsum to ONNX format."""
        equation = params.get("equation")
        input_names = [s.get_name(var) for var in node_inputs]
        output_name = s.get_name(node_outputs[0])

        einsum_node = helper.make_node(
            "Einsum",
            inputs=input_names,
            outputs=[output_name],
            name=s.get_unique_name("einsum"),
            equation=equation,
        )
        s.add_node(einsum_node)

        input_shapes = [inp.aval.shape for inp in node_inputs]
        output_shape = EinsumPlugin._get_dynamic_output_shape(input_shapes, equation)
        s.add_shape_info(output_name, output_shape)

    @staticmethod
    def _einsum(equation, *operands, precision=None):
        """Defines the primitive binding for Einsum. Corrected version."""
        # Pass only the operands as positional arguments, and equation/precision as keywords.
        return jnp.einsum_p.bind(*operands, equation=equation, precision=precision)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for Einsum."""

        def patched_einsum(equation, *operands, precision=None):
            return EinsumPlugin._einsum(equation, *operands, precision=precision)

        return patched_einsum

    @staticmethod
    def patch_info():
        """Provides patching information for Einsum."""
        return {
            "patch_targets": [jnp],
            "patch_function": lambda _: EinsumPlugin.get_monkey_patch(),
            "target_attribute": "einsum",
        }


# Register abstract evaluation function
jnp.einsum_p.def_abstract_eval(EinsumPlugin.abstract_eval)



from jax import core
from jax.extend.core import Primitive
from flax import nnx
from onnx import helper
from typing import TYPE_CHECKING
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the LogSoftmax primitive
nnx.log_softmax_p = Primitive("nnx.log_softmax")
nnx.log_softmax_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=nnx.log_softmax_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.nn.log_softmax.html",
    onnx=[
        {
            "component": "LogSoftmax",
            "doc": "https://onnx.ai/onnx/operators/onnx__LogSoftmax.html",
        }
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "log_softmax",
            "callable": lambda x: nnx.log_softmax(x),
            "input_shapes": [(3,)],
        }
    ],
)
class LogSoftmaxPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.log_softmax to ONNX.
    """

    @staticmethod
    def abstract_eval(x, axis=-1):
        """Abstract evaluation function for LogSoftmax."""
        return core.ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of LogSoftmax to ONNX format."""
        input_var = node_inputs[0]
        output_var = node_outputs[0]

        input_name = s.get_name(input_var)
        output_name = s.get_name(output_var)

        # Retrieve the axis parameter (defaulting to -1 if not provided)
        axis = params.get("axis", -1)

        log_softmax_node = helper.make_node(
            "LogSoftmax",
            inputs=[input_name],
            outputs=[output_name],
            name=s.get_unique_name("log_softmax"),
            axis=axis,
        )
        s.add_node(log_softmax_node)

    @staticmethod
    def _log_softmax(x, axis=-1):
        """Defines the primitive binding for LogSoftmax."""
        return nnx.log_softmax_p.bind(x, axis=axis)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for LogSoftmax."""

        def patched_log_softmax(x, axis=-1):
            return LogSoftmaxPlugin._log_softmax(x, axis)

        return patched_log_softmax

    @staticmethod
    def patch_info():
        """Provides patching information for LogSoftmax."""
        return {
            "patch_targets": [nnx],
            "patch_function": lambda _: LogSoftmaxPlugin.get_monkey_patch(),
            "target_attribute": "log_softmax",
        }


# Register abstract evaluation function
nnx.log_softmax_p.def_abstract_eval(LogSoftmaxPlugin.abstract_eval)



import jax
from typing import TYPE_CHECKING
from onnx import helper
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.sort_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.sort.html",
    onnx=[
        {
            "component": "TopK",
            "doc": "https://onnx.ai/onnx/operators/onnx__TopK.html",
        }
    ],
    since="v0.2.0",
    context="primitives.lax",
    testcases=[
        {
            "testcase": "sort_1d",
            "callable": lambda x: jax.lax.sort(x),
            "input_shapes": [(3,)],
        },
        {
            "testcase": "sort_1d_empty",
            "callable": lambda x: jax.lax.sort(x),
            "input_shapes": [(0,)],  # Empty array
        },
        {
            "testcase": "sort_1d_single",
            "callable": lambda x: jax.lax.sort(x),
            "input_shapes": [(1,)],  # Single-element array
        },
        {
            "testcase": "sort_1d_larger",
            "callable": lambda x: jax.lax.sort(x),
            "input_shapes": [(10,)],
        },
        {
            "testcase": "sort_1d_specific_values",
            "callable": lambda x: jax.lax.sort(x),
            "input_shapes": [(5,)],
            "input_data": [([5, 2, 1, 4, 3],)],  # Specific input data
            "expected_output": [([1, 2, 3, 4, 5],)],  # Expected output
        },
    ],
)
class SortPlugin(PrimitivePlugin):
    """Plugin for converting jax.lax.sort to ONNX TopK."""

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX sort primitive."""
        input_name = s.get_name(node_inputs[0])
        shape_name = s.get_unique_name("sort_shape")
        value_name = s.get_var_name(node_outputs[0])
        indices_name = s.get_unique_name("sort_indices_output")
        if "axis" in params:
            # Not supported for now
            raise NotImplementedError("Sort with axis not supported yet")
        else:
            node_shape = helper.make_node(
                "Shape",
                inputs=[input_name],
                outputs=[shape_name],
                name=s.get_unique_name("shape"),
            )
            s.add_node(node_shape)
        node_topk = helper.make_node(
            "TopK",
            inputs=[input_name, shape_name],
            outputs=[value_name, indices_name],
            name=s.get_unique_name("sort"),
            largest=0,
        )
        s.add_node(node_topk)



import jax
from typing import TYPE_CHECKING
from onnx import helper, TensorProto
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin
import jax.numpy as jnp

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.argmax_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.argmax.html",
    onnx=[
        {
            "component": "ArgMax",
            "doc": "https://onnx.ai/onnx/operators/onnx__ArgMax.html",
        }
    ],
    since="v0.2.0",
    context="primitives.lax",
    testcases=[
        {
            "testcase": "argmax_test1",
            "callable": lambda x: jax.lax.argmax(x, axis=0, index_dtype=jnp.int32),
            "input_shapes": [(3, 3)],
        },
        {
            "testcase": "argmax_test2",
            "callable": lambda x: jax.lax.argmax(x, axis=1, index_dtype=jnp.int32),
            "input_shapes": [(3, 3)],
        },
    ],
)
class ArgMaxPlugin(PrimitivePlugin):
    """
    Plugin for converting jax.lax.argmax to ONNX.
    """

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX argmax primitive."""
        input_name = s.get_name(node_inputs[0])
        intermediate_name = s.get_unique_name("argmax_intermediate")
        output_name = s.get_name(node_outputs[0])  # Corrected: get_name
        axis = params["axes"][0]
        # keepdims is always False for jax.lax.argmax
        keepdims = 0  # Hardcoded: jax.lax.argmax always has keepdims=False

        node_1 = helper.make_node(
            "ArgMax",
            inputs=[input_name],
            outputs=[intermediate_name],
            name=s.get_unique_name("argmax"),
            axis=axis,
            keepdims=keepdims,
            select_last_index=int(
                params["index_dtype"] == jnp.int64
            ),  # Properly set select_last_index, convert bool to int
        )
        s.add_node(node_1)

        node_2 = helper.make_node(
            "Cast",
            inputs=[intermediate_name],
            outputs=[output_name],
            to=TensorProto.INT32,  # Cast to the correct output type (INT32)
        )
        s.add_node(node_2)



from jax import core, numpy as jnp
from jax.extend.core import Primitive
from onnx import helper
from typing import TYPE_CHECKING
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the Add primitive
jnp.add_p = Primitive("jnp.add")
jnp.add_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=jnp.add_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.add.html",
    onnx=[
        {
            "component": "Add",
            "doc": "https://onnx.ai/onnx/operators/onnx__Add.html",
        }
    ],
    since="v0.1.0",
    context="primitives.jnp",
    testcases=[
        {
            "testcase": "add",
            "callable": lambda x, y: jnp.add(x, y),
            "input_shapes": [(3,), (3,)],
        }
    ],
)
class AddPlugin(PrimitivePlugin):
    """
    Plugin for converting jax.numpy.add to ONNX.
    """

    @staticmethod
    def abstract_eval(x, y):
        """Abstract evaluation function for Add."""
        return core.ShapedArray(x.shape, x.dtype)  # Should handle broadcasting

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of Add to ONNX format."""
        # Expect node_inputs: [x, y]
        x_var = node_inputs[0]
        y_var = node_inputs[1]
        output_var = node_outputs[0]

        x_name = s.get_name(x_var)
        y_name = s.get_name(y_var)
        output_name = s.get_name(output_var)

        add_node = helper.make_node(
            "Add",
            inputs=[x_name, y_name],
            outputs=[output_name],
            name=s.get_unique_name("add"),
        )
        s.add_node(add_node)

    @staticmethod
    def _add(x, y):
        """Defines the primitive binding for Add."""
        return jnp.add_p.bind(x, y)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for Add."""

        def patched_add(x, y):
            return AddPlugin._add(x, y)

        return patched_add

    @staticmethod
    def patch_info():
        """Provides patching information for Add."""
        return {
            "patch_targets": [jnp],
            "patch_function": lambda _: AddPlugin.get_monkey_patch(),
            "target_attribute": "add",
        }


# Register abstract evaluation function
jnp.add_p.def_abstract_eval(AddPlugin.abstract_eval)



import jax
from typing import TYPE_CHECKING
from onnx import helper
import jax.numpy as jnp
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.gather_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.gather.html",
    onnx=[
        {
            "component": "Gather",
            "doc": "https://onnx.ai/onnx/operators/onnx__Gather.html",
        }
    ],
    since="v0.2.0",
    context="primitives.lax",
    testcases=[
        {
            "testcase": "gather",
            "callable": lambda x: jax.lax.gather(
                x,
                jnp.array([[1], [0]]),
                jax.lax.GatherDimensionNumbers(
                    offset_dims=(1,),
                    collapsed_slice_dims=(0,),
                    start_index_map=(0,),
                ),
                (1, 3),
            ),
            "input_shapes": [(3, 3)],
        }
    ],
)
class GatherPlugin(PrimitivePlugin):
    """Plugin for converting jax.lax.gather to ONNX Gather."""

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX gather primitive."""
        input_names = [s.get_name(inp) for inp in node_inputs]
        output_name = s.get_var_name(node_outputs[0])

        # Extract parameters from the JAX gather node
        dimension_numbers = params["dimension_numbers"]
        params["slice_sizes"]

        # Determine the axis to gather along based on dimension_numbers
        # This is a simplification; a more robust solution might be required
        # for complex gather operations.
        if len(dimension_numbers.start_index_map) == 1:
            axis = dimension_numbers.start_index_map[0]
        else:
            # Handle cases with multiple start_index_map values (more complex gather)
            # For simplicity, we assume the first start_index_map if multiple exist.
            axis = dimension_numbers.start_index_map[0]

        node = helper.make_node(
            "Gather",  # Use ONNX Gather instead of GatherElements
            inputs=input_names,
            outputs=[output_name],
            name=s.get_unique_name("gather"),
            axis=axis,  # Add the axis parameter.
        )
        s.add_node(node)



import numpy as np
from jax import core
from flax import nnx
from onnx import helper
from jax.extend.core import Primitive
from typing import TYPE_CHECKING, Tuple, Sequence
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the primitive for convolution operations.
nnx.conv_p = Primitive("nnx.conv")
nnx.conv_p.multiple_results = False  # Correctly set at initialization


@register_primitive(
    jaxpr_primitive=nnx.conv_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/linear.html#flax.nnx.Conv",
    onnx=[
        {"component": "Conv", "doc": "https://onnx.ai/onnx/operators/onnx__Conv.html"},
        {
            "component": "Transpose",
            "doc": "https://onnx.ai/onnx/operators/onnx__Transpose.html",
        },
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "conv",
            "callable": nnx.Conv(
                in_features=3,
                out_features=16,
                kernel_size=(3, 3),
                strides=(1, 1),
                padding="SAME",
                use_bias=True,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [("B", 28, 28, 3)],
        },
        {
            "testcase": "conv_2",
            "callable": nnx.Conv(1, 32, kernel_size=(3, 3), rngs=nnx.Rngs(0)),
            "input_shapes": [(2, 28, 28, 1)],
        },
        {
            "testcase": "conv_3",
            "callable": nnx.Conv(
                in_features=1,
                out_features=32,
                kernel_size=(3, 3),
                strides=(1, 1),
                padding="SAME",
                use_bias=True,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(3, 28, 28, 1)],
        },
        {
            "testcase": "conv_4",
            "callable": nnx.Conv(
                in_features=32,
                out_features=64,
                kernel_size=(3, 3),
                strides=(2, 2),
                padding="SAME",
                use_bias=True,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [(3, 28, 28, 32)],
        },
    ],
)
class ConvPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.Conv to ONNX.

    Converts a Conv operation into an ONNX Conv operator with Transpose
    operations to handle format conversion.
    """

    @staticmethod
    def _compute_conv_output_shape(
        x_shape: Tuple[int, ...],
        kernel_shape: Tuple[int, ...],
        strides: Sequence[int] | int,
        padding: str,
    ) -> Tuple[int, ...]:
        """
        Compute the output shape for a 2D convolution.
        Assumes:
          - Input is in NHWC format: (N, H, W, C)
          - Kernel is in HWIO format: (filter_height, filter_width, in_channels, out_channels)
        """
        if isinstance(strides, int):
            strides = (strides, strides)
        N, H, W, _ = x_shape
        filter_height, filter_width, _, out_channels = kernel_shape
        if padding.upper() == "VALID":
            out_H = (H - filter_height) // strides[0] + 1
            out_W = (W - filter_width) // strides[1] + 1
        elif padding.upper() == "SAME":
            # Use ceiling division for SAME padding.
            out_H = -(-H // strides[0])
            out_W = -(-W // strides[1])
        else:
            raise ValueError("Unsupported padding: " + padding)
        return (N, out_H, out_W, out_channels)

    @staticmethod
    def abstract_eval(x, kernel, bias, strides, padding, dilations, dimension_numbers):
        """Abstract evaluation: computes output shape and dtype."""
        out_shape = ConvPlugin._compute_conv_output_shape(
            x.shape, kernel.shape, strides, padding
        )
        return core.ShapedArray(out_shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Convert conv operation to ONNX format."""

        # Expected node_inputs: [input, kernel, (optional) bias]
        input_var = node_inputs[0]
        kernel_var = node_inputs[1]
        bias_var = node_inputs[2] if len(node_inputs) > 2 else None

        # Get names from the converter.
        input_name = s.get_name(input_var)
        final_output_name = s.get_name(node_outputs[0])
        bias_name = s.get_name(bias_var) if bias_var is not None else None

        # Pre-Transpose: Convert input from NHWC -> NCHW.
        pre_transpose_name = s.get_unique_name("pre_transpose")
        pre_transpose_node = helper.make_node(
            "Transpose",
            inputs=[input_name],
            outputs=[pre_transpose_name],
            name=s.get_unique_name("transpose_pre"),
            perm=[0, 3, 1, 2],
        )
        s.add_node(pre_transpose_node)
        # Compute the pre-transposed shape.
        jax_input_shape = input_var.aval.shape  # e.g. (B, H, W, C)
        pre_transposed_shape = tuple(
            jax_input_shape[i] for i in [0, 3, 1, 2]
        )  # (B, C, H, W)
        s.add_shape_info(pre_transpose_name, pre_transposed_shape)

        # Convert kernel constant: from HWIO to OIHW.
        kernel_name = s.get_name(kernel_var)
        kernel_const = s.name_to_const[kernel_name]
        transposed_kernel = np.transpose(kernel_const, [3, 2, 0, 1])
        weights_name = s.get_constant_name(transposed_kernel)

        # Determine convolution parameters.
        strides = params.get("strides", (1, 1))
        if isinstance(strides, int):
            strides = (strides, strides)
        padding = params.get("padding", "VALID")
        dilations = params.get("dilations", (1, 1))

        # Create the Conv node. ONNX Conv expects input in NCHW and kernel in OIHW.
        conv_out_name = s.get_unique_name("conv_output")
        if bias_name is not None:
            conv_node = helper.make_node(
                "Conv",
                inputs=[pre_transpose_name, weights_name, bias_name],
                outputs=[conv_out_name],
                name=s.get_unique_name("conv"),
                strides=strides,
                dilations=dilations,
                pads=[0, 0, 0, 0] if padding.upper() == "VALID" else None,
            )
        else:
            conv_node = helper.make_node(
                "Conv",
                inputs=[pre_transpose_name, weights_name],
                outputs=[conv_out_name],
                name=s.get_unique_name("conv"),
                strides=strides,
                dilations=dilations,
                pads=[0, 0, 0, 0] if padding.upper() == "VALID" else None,
            )
        if padding.upper() == "SAME":
            # Compute symmetric padding for height and width.
            # ONNX expects pads in the order: [pad_top, pad_left, pad_bottom, pad_right]
            input_shape = input_var.aval.shape  # (B, H, W, C)
            filter_shape = kernel_const.shape  # (H, W, I, O)
            # Height padding.
            in_h = input_shape[1]
            filt_h = filter_shape[0]
            stride_h = strides[0]
            out_h = -(-in_h // stride_h)  # Ceiling division
            pad_total_h = max((out_h - 1) * stride_h + filt_h - in_h, 0)
            pad_top = pad_total_h // 2
            pad_bottom = pad_total_h - pad_top
            # Width padding.
            in_w = input_shape[2]
            filt_w = filter_shape[1]
            stride_w = strides[1]
            out_w = -(-in_w // stride_w)
            pad_total_w = max((out_w - 1) * stride_w + filt_w - in_w, 0)
            pad_left = pad_total_w // 2
            pad_right = pad_total_w - pad_left
            pads = [pad_top, pad_left, pad_bottom, pad_right]
            conv_node.attribute.append(helper.make_attribute("pads", pads))
        s.add_node(conv_node)
        # Compute the conv node's intermediate output shape (in NCHW):
        # First, get the expected final output shape in JAX (NHWC) using our helper:
        jax_output_shape = ConvPlugin._compute_conv_output_shape(
            jax_input_shape, kernel_const.shape, strides, padding
        )
        # Then, compute the intermediate shape by transposing NHWC -> NCHW.
        conv_output_shape_NCHW = tuple(jax_output_shape[i] for i in [0, 3, 1, 2])
        s.add_shape_info(conv_out_name, conv_output_shape_NCHW)

        # Post-Transpose: Convert Conv output from NCHW -> NHWC.
        post_transpose_node = helper.make_node(
            "Transpose",
            inputs=[conv_out_name],
            outputs=[final_output_name],
            name=s.get_unique_name("transpose_post"),
            perm=[0, 2, 3, 1],
        )
        s.add_node(post_transpose_node)
        # The final output shape should match the JAX output shape.
        # s.add_shape_info(final_output_name, jax_output_shape)

    @staticmethod
    def _conv(x, kernel, bias, strides, padding, dilations, dimension_numbers):
        return nnx.conv_p.bind(
            x,
            kernel,
            bias,
            strides=strides,
            padding=padding,
            dilations=dilations,
            dimension_numbers=dimension_numbers,
        )

    @staticmethod
    def get_monkey_patch():
        """Returns a patched version of Conv.__call__."""

        def patched_conv_call(self, x):
            return ConvPlugin._conv(
                x,
                self.kernel.value,
                self.bias.value if self.bias else None,
                self.strides,
                self.padding,
                getattr(self, "dilations", (1, 1)),
                getattr(self, "dimension_numbers", None),
            )

        return patched_conv_call

    @staticmethod
    def patch_info():
        return {
            "patch_targets": [nnx.Conv],
            "patch_function": lambda _: ConvPlugin.get_monkey_patch(),
        }


# Register abstract evaluation function.
nnx.conv_p.def_abstract_eval(ConvPlugin.abstract_eval)



import numpy as np
import onnxruntime as ort


def allclose(callable, onnx_model_path, *xs):
    # Test ONNX and JAX outputs
    session = ort.InferenceSession(onnx_model_path)

    # Prepare inputs for ONNX model
    p = {"var_" + str(i): np.array(x) for i, x in enumerate(xs)}
    onnx_output = session.run(None, p)

    # Get JAX model output
    jax_output = callable(*xs)

    # Verify outputs match
    if not isinstance(jax_output, list):
        jax_output = [jax_output]
    if not isinstance(onnx_output, list):
        onnx_output = [onnx_output]

    isOk = np.allclose(onnx_output, jax_output, rtol=1e-3, atol=1e-5)

    return (
        isOk,
        (
            "ONNX and JAX outputs match :-)"
            if isOk
            else "ONNX and JAX outputs do not match :-("
        ),
    )



import pkgutil
import importlib
import os
from typing import Optional, Callable, Dict, Any, Type, Union

PLUGIN_REGISTRY: Dict[str, Union["ExamplePlugin", "PrimitivePlugin"]] = {}


class PrimitivePlugin:
    """Base class for ONNX conversion plugins."""

    primitive: str
    metadata: Dict[str, Any]
    patch_info: Optional[Callable[[], Dict[str, Any]]] = (
        None  # Method returning patch details
    )

    def get_handler(self, converter: Any) -> Callable:
        return lambda node_inputs, node_outputs, params: self.to_onnx(
            converter, node_inputs, node_outputs, params
        )

    def to_onnx(
        self, converter: Any, node_inputs: Any, node_outputs: Any, params: Any
    ) -> None:
        """Handles JAX to ONNX conversion; must be overridden."""
        raise NotImplementedError


class ExamplePlugin:
    metadata: Dict[str, Any]


def register_example(**metadata: Any) -> ExamplePlugin:
    """
    Decorator for registering an example plugin.
    The metadata must be a dictionary of attributes.
    """
    instance = ExamplePlugin()
    instance.metadata = metadata
    component = metadata.get("component")
    if isinstance(component, str):
        PLUGIN_REGISTRY[component] = instance
    return instance


def register_primitive(
    **metadata: Any,
) -> Callable[[Type[PrimitivePlugin]], Type[PrimitivePlugin]]:
    """
    Decorator to register a plugin with the given primitive and metadata.
    """
    primitive = metadata.get("jaxpr_primitive", "")

    def decorator(cls: Type[PrimitivePlugin]) -> Type[PrimitivePlugin]:
        if not issubclass(cls, PrimitivePlugin):
            raise TypeError("Plugin must subclass PrimitivePlugin")

        instance = cls()
        instance.primitive = primitive
        instance.metadata = metadata or {}

        # Register patch_info if defined in the class
        if hasattr(cls, "patch_info"):
            instance.patch_info = getattr(cls, "patch_info")

        if isinstance(primitive, str):
            PLUGIN_REGISTRY[primitive] = instance
        return cls

    return decorator


_already_imported_plugins = False


def import_all_plugins() -> None:
    """Imports all plugins dynamically from the 'plugins' directory."""
    global _already_imported_plugins
    if _already_imported_plugins:
        return  # Already imported plugins; no-op
    plugins_path = os.path.join(os.path.dirname(__file__), "plugins")
    for _, module_name, _ in pkgutil.walk_packages(
        [plugins_path], prefix="jax2onnx.plugins."
    ):
        importlib.import_module(module_name)
    _already_imported_plugins = True  # Mark as imported



# jax2onnx/converter/converter.py

import jax
import onnx
from onnx import helper
import numpy as np
from typing import Dict, Any, Tuple
import jax.random
import contextlib
import inspect
from jax2onnx.converter.onnx_builder import OnnxBuilder
from jax2onnx.converter.optimize_onnx_graph import (
    remove_redundant_transpose_pairs,
    remove_redundant_casts,
)
from jax2onnx.plugin_system import (
    PLUGIN_REGISTRY,
    PrimitivePlugin,
    import_all_plugins,
)


def save_onnx(
    fn: Any,
    input_shapes: Any,
    output_path: str = "model.onnx",
    model_name: str = "jax_model",
    include_intermediate_shapes: bool = True,
    opset: int = 21,
) -> str:
    import_all_plugins()
    jaxpr2onnx = JaxprToOnnx()
    return jaxpr2onnx.save_onnx(
        fn,
        input_shapes,
        output_path=output_path,
        model_name=model_name,
        include_intermediate_shapes=include_intermediate_shapes,
        opset=opset,
    )


class JaxprToOnnx:
    def save_onnx(
        self,
        fn: Any,
        input_shapes: Any,
        output_path: str = "model.onnx",
        model_name: str = "jax_model",
        include_intermediate_shapes: bool = True,
        opset: int = 21,
    ) -> str:

        # if input_shapes have dynamic batch dimensions then include_intermediate_shapes must be False
        if any("B" in shape for shape in input_shapes):
            include_intermediate_shapes = False
            print(
                "Dynamic batch dimensions detected. Setting include_intermediate_shapes=False"
            )

        self._validate_input_shapes(input_shapes=input_shapes)
        example_args = [
            jax.numpy.zeros(self._shape_with_example_batch(s)) for s in input_shapes
        ]

        # example_args = create_example_args_with_dynamic_batch(input_shapes)

        with temporary_monkey_patches():
            jax.make_jaxpr(fn)(*example_args)

        converter = Jaxpr2OnnxConverter()
        converter.trace_jaxpr(fn, example_args)

        # Set symbolic batch dimension 'B' only on corresponding input tensors
        for tensor, input_shape in zip(converter.builder.inputs, input_shapes):
            tensor_shape = tensor.type.tensor_type.shape.dim
            for idx, dim in enumerate(input_shape):
                if dim == "B":
                    tensor_shape[idx].dim_param = "B"

        # Set symbolic batch dimension 'B' on outputs if it's set on any input
        batch_dims = {
            idx for shape in input_shapes for idx, dim in enumerate(shape) if dim == "B"
        }
        for tensor in converter.builder.outputs:
            tensor_shape = tensor.type.tensor_type.shape.dim
            for idx in batch_dims:
                if idx < len(tensor_shape):
                    tensor_shape[idx].dim_param = "B"

        # Optionally include intermediate shape information
        value_info = converter.builder.value_info if include_intermediate_shapes else []

        # Remove unused initializers
        used_initializers = {i for node in converter.builder.nodes for i in node.input}
        converter.builder.initializers = [
            init
            for init in converter.builder.initializers
            if init.name in used_initializers
        ]

        graph = helper.make_graph(
            nodes=converter.builder.nodes,
            name=model_name,
            inputs=converter.builder.inputs,
            outputs=converter.builder.outputs,
            initializer=converter.builder.initializers,
            value_info=value_info,
        )

        onnx_model = helper.make_model(
            graph, opset_imports=[helper.make_opsetid("", opset)]
        )

        onnx_model = remove_redundant_transpose_pairs(
            remove_redundant_casts(onnx_model)
        )

        onnx.save_model(onnx_model, output_path)

        return output_path

    def _validate_input_shapes(self, input_shapes):
        for shape in input_shapes:
            assert isinstance(shape, tuple), "Each input shape must be a tuple"

    def _shape_with_example_batch(self, shape, example_batch=2):
        return tuple(example_batch if d == "B" else d for d in shape)


class Jaxpr2OnnxConverter:
    """
    A translator that converts JAX's JAXPR representation to ONNX format.
    """

    def __init__(self, name_counter=0):

        # Instead of duplicating helper functions, delegate to OnnxBuilder:
        self.builder = OnnxBuilder(name_counter)
        # Other converter state
        self.var_to_name: Dict[Any, str] = {}
        self.name_to_var: Dict[str, Any] = {}
        self.primitive_handlers = {}
        # for primitive, plugin in get_all_plugins().items():
        #     handler = plugin.get_handler(self)
        #     self.primitive_handlers[primitive] = handler
        self.primitive_handlers[jax._src.prng.random_seed_p] = self._handle_random_seed
        self.primitive_handlers[jax._src.prng.random_wrap_p] = self._handle_random_wrap
        self.primitive_handlers[jax._src.prng.random_split_p] = (
            self._handle_random_split
        )
        self.primitive_handlers[jax._src.prng.random_unwrap_p] = (
            self._handle_random_unwrap
        )

        import_all_plugins()

        for key in PLUGIN_REGISTRY:
            plugin = PLUGIN_REGISTRY[key]
            if isinstance(plugin, PrimitivePlugin):
                self.primitive_handlers[key] = plugin.get_handler(self)

    def new_var(self, dtype: np.dtype, shape: Tuple[int, ...]):
        return jax.core.Var(
            self.builder.get_unique_name(""), jax.core.ShapedArray(shape, dtype)
        )

    def add_node(self, node):
        self.builder.add_node(node)

    def get_unique_name(self, prefix="node"):
        return self.builder.get_unique_name(prefix)

    def get_var_name(self, var):
        if var not in self.var_to_name:
            self.var_to_name[var] = self.get_unique_name("var")
        return self.var_to_name[var]

    def get_constant_name(self, val):
        return self.builder.get_constant_name(val)

    def add_input(self, var, shape, dtype=np.float32):
        name = self.get_var_name(var)
        self.builder.add_input(name, shape, dtype)
        return name

    def add_output(self, var, shape, dtype=np.float32):
        name = self.get_var_name(var)
        self.builder.add_output(name, shape, dtype)
        return name

    def add_shape_info(self, name, shape, dtype=np.float32):
        self.builder.add_value_info(name, shape, dtype)
        return name

    def get_name(self, var):
        if isinstance(var, jax._src.core.Var):
            return self.get_var_name(var)
        elif isinstance(var, jax._src.core.Literal):
            return self.get_constant_name(var)
        else:
            raise NotImplementedError("not yet implemented")

    def finalize_model(self, output_path, model_name):
        graph = self.builder.create_graph(model_name)
        onnx_model = self.builder.create_model(graph)
        onnx.save_model(onnx_model, output_path)
        return output_path

    def _create_identity_node(self, node_inputs, node_outputs, name_prefix):
        """Create an Identity node connecting inputs to outputs."""
        input_names = [self.get_name(inp) for inp in node_inputs]
        output_name = self.get_var_name(node_outputs[0])

        node = self.builder.create_node(
            "Identity",
            input_names,
            [output_name],
            name=self.get_unique_name(name_prefix),
        )
        self.builder.add_node(node)
        return output_name

    def _handle_random_seed(self, node_inputs, node_outputs, params):
        return self._create_identity_node(node_inputs, node_outputs, "random_seed")

    def _handle_random_wrap(self, node_inputs, node_outputs, params):
        return self._create_identity_node(node_inputs, node_outputs, "random_wrap")

    def _handle_random_unwrap(self, node_inputs, node_outputs, params):
        return self._create_identity_node(node_inputs, node_outputs, "random_unwrap")

    def _handle_random_split(self, node_inputs, node_outputs, params):
        input_name = self._get_name(node_inputs[0])
        intermediate = self._get_unique_name("random_split:x")
        output_name = self._get_var_name(node_outputs[0])

        reshape = self._get_constant_name(np.array([1, 2], dtype=np.int64))

        num = params["shape"][0]
        repeat = self._get_constant_name(np.array([num, 1], dtype=np.int64))

        node_1 = helper.make_node(
            "Reshape",
            inputs=[input_name, reshape],
            outputs=[intermediate],
            name=self._get_unique_name("random_split:reshape"),
        )
        self.nodes.append(node_1)

        node_2 = helper.make_node(
            "Tile",
            inputs=[intermediate, repeat],
            outputs=[output_name],
            name=self._get_unique_name("random_split:tile"),
        )
        self.nodes.append(node_2)

    def _create_random_distribution_node(self, node_outputs, op_type, name_prefix):
        """Create a node for random distribution operations."""
        output_name = self.get_var_name(node_outputs[0])
        shape = node_outputs[0].aval.shape
        if shape == ():
            shape = (1,)
        node = self.builder.create_node(
            op_type,
            [],
            [output_name],
            name=self.get_unique_name(name_prefix),
            shape=shape,
        )
        self.builder.add_node(node)
        return output_name

    def _handle_random_uniform(self, node_inputs, node_outputs, params):
        return self._create_random_distribution_node(
            node_outputs, "RandomUniform", "random_uniform"
        )

    def _handle_random_normal(self, node_inputs, node_outputs, params):
        return self._create_random_distribution_node(
            node_outputs, "RandomNormal", "random_normal"
        )

    def _handle_truncated_normal(self, node_inputs, node_outputs, params):
        return self._create_random_distribution_node(
            node_outputs, "RandomNormal", "truncated_normal"
        )

    def _handle_convert_element_type(self, node_inputs, node_outputs, params):
        input_names = [self.get_name(inp) for inp in node_inputs]
        output_name = self.get_var_name(node_outputs[0])

        new_dtype = self.builder.numpy_dtype_to_onnx(params["new_dtype"])
        node = self.builder.create_node(
            "Cast",
            input_names,
            [output_name],
            name=self.get_unique_name("convert_element_type"),
            to=new_dtype,
        )
        self.builder.add_node(node)

    def _handle_device_put(self, node_inputs, node_outputs, params):
        name = self.get_unique_name("const")
        # Convert to numpy and create tensor
        val = node_inputs[0]
        actual_val = val.val

        np_val = np.array(actual_val)
        if np_val.dtype == np.int64:
            np_val = np_val.astype(np.int32)
        elif np_val.dtype == np.float64:
            np_val = np_val.astype(np.float32)

        tensor = self.builder.create_tensor(
            name=name,
            data_type=self.builder.numpy_dtype_to_onnx(np_val.dtype),
            dims=np_val.shape,
            vals=np_val.flatten().tolist(),
        )
        self.builder.add_initializer(tensor)
        input_names = [name]
        output_name = self.get_var_name(node_outputs[0])

        node = self.builder.create_node(
            "Identity",
            input_names,
            [output_name],
            name=self.get_unique_name("device_put"),
        )
        self.builder.add_node(node)

    def _process_pjit(self, jaxpr):
        closed_jaxpr = jaxpr.params["jaxpr"]
        if not isinstance(closed_jaxpr, jax._src.core.ClosedJaxpr):
            raise ValueError("Expected ClosedJaxpr in pjit.param[jaxpr]")

        name = jaxpr.params["name"]

        # Group related operations by handler type
        random_distributions = {
            "_normal": self._handle_random_normal,
            "_uniform": self._handle_random_uniform,
            "_truncated_normal": self._handle_truncated_normal,
        }

        closed_jaxpr_ops = ["_gamma", "clip", "sort", "_where", "_gumbel", "_dirichlet"]

        if name in random_distributions:
            random_distributions[name](jaxpr.invars, jaxpr.outvars, jaxpr.params)
        elif name in closed_jaxpr_ops:
            self._process_closed_jaxpr(jaxpr)
        else:
            raise NotImplementedError(f"pjit {name} not yet handled")

    def _process_eqn(self, jaxpr):
        """Process a single JAXPR equation."""
        if hasattr(jaxpr, "primitive"):
            primitive = jaxpr.primitive
            if primitive.name == "pjit":
                self._process_pjit(jaxpr)
            elif primitive.name in self.primitive_handlers:

                self.primitive_handlers[primitive.name](
                    jaxpr.invars, jaxpr.outvars, jaxpr.params
                )
                # for all jaxpr.outvars, add shape info
                for outvar in jaxpr.outvars:
                    output_name = self.get_name(outvar)
                    self.add_shape_info(output_name, outvar.aval.shape)

            else:
                raise NotImplementedError(f"Primitive {primitive.name} not implemented")
        else:
            # Handle call primitives or other special cases
            raise NotImplementedError(f"Non-primitive equation: {jaxpr}")

    def _process_closed_jaxpr(self, jaxpr):
        """Process a closed JAXPR inside a JaxprEqn."""
        assert isinstance(jaxpr, jax._src.core.JaxprEqn)

        closed_jaxpr = jaxpr.params["jaxpr"]
        node_inputs = jaxpr.invars
        node_outputs = jaxpr.outvars

        # Create a subconverter to process the nested jaxpr
        subconverter = Jaxpr2OnnxConverter(self.builder.name_counter + 1)
        subconverter._process_jaxpr(closed_jaxpr.jaxpr, closed_jaxpr.consts)

        # Connect inputs from parent to subconverter
        self._connect_inputs_to_subconverter(node_inputs, subconverter.builder.inputs)

        # Add the subconverter's nodes and initializers to our own
        self.builder.nodes.extend(subconverter.builder.nodes)
        self.builder.initializers.extend(subconverter.builder.initializers)
        self.builder.name_counter += (
            subconverter.builder.name_counter - subconverter.builder.name_counter_init
        )

        # Connect outputs from subconverter back to parent
        self._connect_outputs_from_subconverter(
            node_outputs, subconverter.builder.outputs
        )

    def _connect_inputs_to_subconverter(self, parent_inputs, subconverter_inputs):
        """Connect inputs from parent to subconverter."""
        assert len(parent_inputs) == len(subconverter_inputs)

        for parent_input, subconverter_input in zip(parent_inputs, subconverter_inputs):
            parent_name = self.get_name(parent_input)
            subconverter_name = subconverter_input.name
            node = self.builder.create_node(
                "Identity",
                [parent_name],
                [subconverter_name],
                name=self.get_unique_name("pjit_input"),
            )
            self.builder.add_node(node)

    def _connect_outputs_from_subconverter(self, parent_outputs, subconverter_outputs):
        """Connect outputs from subconverter back to parent."""
        assert len(parent_outputs) == len(subconverter_outputs)

        for parent_output, subconverter_output in zip(
            parent_outputs, subconverter_outputs
        ):
            parent_name = self.get_name(parent_output)
            subconverter_name = subconverter_output.name
            node = self.builder.create_node(
                "Identity",
                [subconverter_name],
                [parent_name],
                name=self.get_unique_name("pjit_output"),
            )
            self.builder.add_node(node)

    def _process_jaxpr(self, jaxpr, consts):
        """Process a JAXPR and convert it to ONNX nodes."""

        # Setup inputs
        for var in jaxpr.invars:
            self.add_input(var, var.aval.shape, var.aval.dtype)

        # Setup constants
        for i, const in enumerate(consts):
            const_name = self.get_constant_name(const)
            const_var = jaxpr.constvars[i]
            self.var_to_name[const_var] = const_name
            self.name_to_const[const_name] = const

        # Process all equations in the JAXPR
        for eqn in jaxpr.eqns:
            self._process_eqn(eqn)

        # Setup outputs
        for var in jaxpr.outvars:
            self.add_output(var, var.aval.shape, var.aval.dtype)

    def trace_jaxpr(self, fn, example_args):
        # Reset state
        self.builder.reset()
        self.var_to_name = {}
        self.name_to_const = {}

        # Get JAXPR from the function
        with temporary_monkey_patches():
            closed_jaxpr = jax.make_jaxpr(fn)(*example_args)

        print(closed_jaxpr)

        jaxpr, consts = closed_jaxpr.jaxpr, closed_jaxpr.consts
        self._process_jaxpr(jaxpr, consts)

    def convert(
        self, fn, example_args, output_path="model.onnx", model_name="jax_model"
    ):
        """
        Convert a JAX function to ONNX.

        Args:
            fn: JAX function to convert
            example_args: Example input arguments to trace the function
            output_path: Path to save the ONNX model

        Returns:
            Path to the saved ONNX model
        """

        self.trace_jaxpr(fn, example_args)

        # Remove unused initializers
        used_initializers = {i for node in self.builder.nodes for i in node.input}
        self.builder.initializers = [
            init for init in self.builder.initializers if init.name in used_initializers
        ]

        graph = self.builder.create_graph(model_name)

        # Create ONNX model
        onnx_model = self.builder.create_model(graph)

        # Save model
        onnx.save_model(onnx_model, output_path)
        return output_path

    def add_initializer(
        self, name, vals, data_type=helper.TensorProto.INT64, dims=None
    ):
        """Add a tensor initializer to the model.

        Args:
            name: The name of the initializer
            vals: The values to initialize with
            data_type: The data type of the tensor (default: INT64)
            dims: The dimensions of the tensor (default: [len(vals)])

        Returns:
            The name of the created initializer
        """
        if dims is None:
            dims = [len(vals)]

        tensor = helper.make_tensor(
            name=name,
            data_type=data_type,
            dims=dims,
            vals=vals,
        )
        self.builder.initializers.append(tensor)
        return name


@contextlib.contextmanager
def temporary_monkey_patches():
    with contextlib.ExitStack() as stack:
        for key in PLUGIN_REGISTRY:
            plugin = PLUGIN_REGISTRY[key]
            if not isinstance(plugin, PrimitivePlugin) or not plugin.patch_info:
                continue
            patch_info = plugin.patch_info()

            target = patch_info["patch_targets"][0]
            patch_func = patch_info["patch_function"]
            attr = patch_info.get("target_attribute", "__call__")
            stack.enter_context(_temporary_patch(target, attr, patch_func))
        yield


@contextlib.contextmanager
def _temporary_patch(target, attr, patch_func):
    original = getattr(target, attr)

    # Check if the patch function expects an argument
    if inspect.signature(patch_func).parameters:
        patched = patch_func(original)
    else:
        patched = patch_func()  # Call without arguments if none are expected

    setattr(target, attr, patched)
    try:
        yield
    finally:
        setattr(target, attr, original)



# file: quickstart.py

from jax2onnx import save_onnx
from flax import nnx


# Example: A minimal MLP (from Flax documentation)
class MLP(nnx.Module):
    def __init__(self, din, dmid, dout, *, rngs):
        self.linear1 = nnx.Linear(din, dmid, rngs=rngs)
        self.dropout = nnx.Dropout(rate=0.1, rngs=rngs)
        self.bn = nnx.BatchNorm(dmid, rngs=rngs)
        self.linear2 = nnx.Linear(dmid, dout, rngs=rngs)

    def __call__(self, x):
        x = nnx.gelu(self.dropout(self.bn(self.linear1(x))))
        return self.linear2(x)


# Instantiate model
my_callable = MLP(din=30, dmid=20, dout=10, rngs=nnx.Rngs(0))

# Convert and save to ONNX
save_onnx(
    my_callable,
    [("B", 30)],  # Input shapes, batch size 'B' is symbolic
    "my_callable.onnx",  # Output path
)



# file: jax2onnx/converter/onnx_builder.py

from onnx import helper, TensorProto, NodeProto, ValueInfoProto, ModelProto, GraphProto
import numpy as np
from typing import List, Any, Tuple
from jax.extend.core import Literal


class OnnxBuilder:
    def __init__(self, name_counter: int = 0, opset_version: int = 21) -> None:
        self.nodes: List[NodeProto] = []
        self.inputs: List[ValueInfoProto] = []
        self.outputs: List[ValueInfoProto] = []
        self.initializers: List[Any] = []  # TensorProto objects
        self.value_info: List[ValueInfoProto] = []
        self.name_counter: int = name_counter
        self.opset_version: int = opset_version

    def reset(self) -> None:
        self.nodes = []
        self.inputs = []
        self.outputs = []
        self.initializers = []
        self.value_info = []
        self.name_counter = 0

    def get_unique_name(self, prefix: str = "node") -> str:
        name = f"{prefix}_{self.name_counter}"
        self.name_counter += 1
        return name

    def get_constant_name(self, val):
        name = self.get_unique_name("const")
        # Unwrap a JAX Literal to get its Python value.

        if isinstance(val, Literal):
            val = val.val
        # Continue with conversion, for example converting to numpy if needed:
        np_val = np.array(val)
        if np_val.dtype == np.float64:
            np_val = np_val.astype(np.float32)
        tensor = helper.make_tensor(
            name=name,
            data_type=self._numpy_dtype_to_onnx(np_val.dtype),
            dims=np_val.shape,
            vals=np_val.flatten().tolist(),
        )
        self.initializers.append(tensor)
        return name

    def add_input(
        self, name: str, shape: Tuple[int, ...], dtype: Any = np.float32
    ) -> None:
        input_def = helper.make_tensor_value_info(
            name, self._numpy_dtype_to_onnx(dtype), shape
        )
        self.inputs.append(input_def)

    def add_output(
        self, name: str, shape: Tuple[int, ...], dtype: Any = np.float32
    ) -> None:
        output_def = helper.make_tensor_value_info(
            name, self._numpy_dtype_to_onnx(dtype), shape
        )
        self.outputs.append(output_def)

    def add_value_info(
        self, name: str, shape: Tuple[int, ...], dtype: Any = np.float32
    ) -> None:
        value_info = helper.make_tensor_value_info(
            name, self._numpy_dtype_to_onnx(dtype), shape
        )
        self.value_info.append(value_info)

    def create_node(
        self, op_type: str, inputs: List[str], outputs: List[str], **kwargs: Any
    ) -> NodeProto:
        return helper.make_node(op_type, inputs, outputs, **kwargs)

    def add_node(self, node: NodeProto) -> None:
        self.nodes.append(node)

    def create_graph(self, name: str) -> GraphProto:
        return helper.make_graph(
            nodes=self.nodes,
            name=name,
            inputs=self.inputs,
            outputs=self.outputs,
            initializer=self.initializers,
            value_info=self.value_info,
        )

    def create_model(self, graph: GraphProto) -> ModelProto:
        return helper.make_model(
            graph, opset_imports=[helper.make_opsetid("", self.opset_version)]
        )

    def _numpy_dtype_to_onnx(self, dtype: Any) -> int:
        if dtype == np.float32:
            return TensorProto.FLOAT
        elif dtype == np.float64:
            return TensorProto.DOUBLE
        elif dtype == np.int32:
            return TensorProto.INT32
        elif dtype == np.int64:
            return TensorProto.INT64
        elif dtype == np.bool_:
            return TensorProto.BOOL
        else:
            return TensorProto.FLOAT



"""
Group Norm Plugin for JAX to ONNX conversion.

This plugin enables conversion of flax.nnx.GroupNorm layers to ONNX format.
It transforms JAX’s group_norm operations into an ONNX GroupNormalization operator
with necessary Transpose operations for NHWC to NCHW conversion.
"""

from typing import TYPE_CHECKING
from flax import nnx
from jax import core
from onnx import helper
from jax.extend.core import Primitive
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define a new primitive for group norm.
nnx.group_norm_p = Primitive("nnx.group_norm")
nnx.group_norm_p.multiple_results = False  # Set at initialization


@register_primitive(
    jaxpr_primitive=nnx.group_norm_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/normalization.html#flax.nnx.GroupNorm",
    onnx=[
        {
            "component": "GroupNormalization",
            "doc": "https://example.com/onnx_GroupNormalization_doc",  # Replace with an appropriate doc link if available.
        },
    ],
    since="v0.3.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "group_norm",
            "callable": nnx.GroupNorm(num_features=64, rngs=nnx.Rngs(0)),
            "input_shapes": [(11, 2, 2, 64)],
        },
        {
            "testcase": "group_norm_2",
            "callable": nnx.GroupNorm(num_features=16, num_groups=4, rngs=nnx.Rngs(0)),
            "input_shapes": [(2, 16)],
        },
    ],
)
class GroupNormPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.GroupNorm to ONNX.

    Converts a GroupNorm operation into a GroupNormalization operator
    with necessary Transpose operations for NHWC to NCHW conversion.
    """

    @staticmethod
    def abstract_eval(x, scale, bias, *args, **kwargs):
        """Abstract evaluation function for group_norm."""
        return core.ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of group_norm to ONNX format."""
        input_name = s.get_name(node_inputs[0])
        scale_name = s.get_name(node_inputs[1])
        bias_name = s.get_name(node_inputs[2])
        final_output_name = s.get_name(node_outputs[0])
        epsilon = params.get("epsilon", 1e-5)
        num_groups = params.get("num_groups")

        jax_shape = node_inputs[0].aval.shape  # e.g. (11, 2, 2, 64) or (2, 20)

        if len(jax_shape) == 4:
            pre_transpose_name = s.get_unique_name("gn_pre_transpose")
            pre_transpose_node = helper.make_node(
                "Transpose",
                inputs=[input_name],
                outputs=[pre_transpose_name],
                name=s.get_unique_name("gn_transpose_pre"),
                perm=[0, 3, 1, 2],  # NHWC -> NCHW
            )
            s.add_node(pre_transpose_node)
            pre_transposed_shape = (
                jax_shape[0],
                jax_shape[3],
                jax_shape[1],
                jax_shape[2],
            )
            s.add_shape_info(pre_transpose_name, pre_transposed_shape)

            gn_output_name = s.get_unique_name("gn_output")
            group_norm_node = helper.make_node(
                "GroupNormalization",
                inputs=[pre_transpose_name, scale_name, bias_name],
                outputs=[gn_output_name],
                name=s.get_unique_name("group_norm"),
                epsilon=epsilon,
                num_groups=num_groups,
            )
            s.add_node(group_norm_node)
            s.add_shape_info(gn_output_name, pre_transposed_shape)

            post_transpose_node = helper.make_node(
                "Transpose",
                inputs=[gn_output_name],
                outputs=[final_output_name],
                name=s.get_unique_name("gn_transpose_post"),
                perm=[0, 2, 3, 1],  # NCHW -> NHWC
            )
            s.add_node(post_transpose_node)
        else:
            group_norm_node = helper.make_node(
                "GroupNormalization",
                inputs=[input_name, scale_name, bias_name],
                outputs=[final_output_name],
                name=s.get_unique_name("group_norm"),
                epsilon=epsilon,
                num_groups=num_groups,
            )
            s.add_node(group_norm_node)

    @staticmethod
    def _group_norm(x, scale, bias, epsilon, num_groups):
        nnx.group_norm_p.multiple_results = False
        return nnx.group_norm_p.bind(
            x, scale, bias, epsilon=epsilon, num_groups=num_groups
        )

    @staticmethod
    def group_norm(x, scale, bias, epsilon, num_groups):
        """Binding function for group_norm."""
        return GroupNormPlugin._group_norm(x, scale, bias, epsilon, num_groups)

    @staticmethod
    def get_monkey_patch():
        """Returns a patched version of GroupNorm.__call__."""

        def patched_group_norm_call(self, x):
            return GroupNormPlugin._group_norm(
                x,
                self.scale.value,
                self.bias.value,
                epsilon=self.epsilon,
                num_groups=self.num_groups,
            )

        return patched_group_norm_call

    @staticmethod
    def patch_info():
        """Provides patching information."""
        return {
            "patch_targets": [nnx.GroupNorm],
            "patch_function": lambda _: GroupNormPlugin.get_monkey_patch(),
            "target_attribute": "__call__",
        }


# Register abstract evaluation function.
nnx.group_norm_p.def_abstract_eval(GroupNormPlugin.abstract_eval)



"""
RMS Norm Plugin for JAX to ONNX conversion.

This plugin enables conversion of flax.nnx.RMSNorm layers to ONNX format.
It transforms JAX’s rms_norm operations into an ONNX RMSNormalization operator
with necessary Transpose operations for NHWC to NCHW conversion.
"""

from typing import TYPE_CHECKING
from flax import nnx
from jax import core
from onnx import helper
from jax.extend.core import Primitive
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin
import numpy as np

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define a new primitive for RMS norm.
nnx.rms_norm_p = Primitive("nnx.rms_norm")
nnx.rms_norm_p.multiple_results = False  # Set at initialization


@register_primitive(
    jaxpr_primitive=nnx.rms_norm_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/normalization.html#flax.nnx.RMSNorm",
    onnx=[
        {
            "component": "RMSNormalization",
            "doc": "https://example.com/onnx_RMSNormalization_doc",  # Replace with an appropriate doc link if available.
        },
    ],
    since="v0.3.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "rms_norm",
            "callable": nnx.RMSNorm(6, rngs=nnx.Rngs(0)),
            "input_shapes": [(11, 2, 2, 6)],
        },
        {
            "testcase": "rms_norm_2",
            "callable": nnx.RMSNorm(num_features=20, rngs=nnx.Rngs(0)),
            "input_shapes": [(2, 20)],
        },
    ],
)
class RMSNormPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.RMSNorm to ONNX.

    Converts an RMSNorm operation into an RMSNormalization operator
    with necessary Transpose operations for NHWC to NCHW conversion.
    """

    @staticmethod
    def abstract_eval(x, scale, *args, **kwargs):
        """Abstract evaluation function for rms_norm."""
        return core.ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of rms_norm to ONNX format."""
        input_name = s.get_name(node_inputs[0])
        scale_name = s.get_name(node_inputs[1])
        final_output_name = s.get_name(node_outputs[0])
        epsilon = params.get("epsilon", 1e-5)

        node_inputs[0].aval.shape  # e.g. (11, 2, 2, 64) or (2, 20)

        # Calculate mean of squares
        mean_square_name = s.get_unique_name("mean_square")
        mean_square_node = helper.make_node(
            "ReduceMean",
            inputs=[input_name],
            outputs=[mean_square_name],
            name=s.get_unique_name("reduce_mean_square"),
            keepdims=1,  # Change to 1
        )
        s.add_node(mean_square_node)

        # Subtract mean square from input
        sub_name = s.get_unique_name("sub")
        sub_node = helper.make_node(
            "Sub",
            inputs=[input_name, mean_square_name],
            outputs=[sub_name],
            name=s.get_unique_name("sub_mean_square"),
        )
        s.add_node(sub_node)

        # Square the result
        square_name = s.get_unique_name("square")
        square_node = helper.make_node(
            "Pow",
            inputs=[sub_name, s.get_constant_name(np.array(2.0, dtype=np.float32))],
            outputs=[square_name],
            name=s.get_unique_name("square"),
        )
        s.add_node(square_node)

        # Calculate mean of squares
        mean_square_2_name = s.get_unique_name("mean_square_2")
        mean_square_2_node = helper.make_node(
            "ReduceMean",
            inputs=[square_name],
            outputs=[mean_square_2_name],
            name=s.get_unique_name("reduce_mean_square_2"),
            keepdims=1,  # Change to 1
        )
        s.add_node(mean_square_2_node)

        # Add epsilon
        add_epsilon_name = s.get_unique_name("add_epsilon")
        add_epsilon_node = helper.make_node(
            "Add",
            inputs=[
                mean_square_2_name,
                s.get_constant_name(np.array(epsilon, dtype=np.float32)),
            ],
            outputs=[add_epsilon_name],
            name=s.get_unique_name("add_epsilon"),
        )
        s.add_node(add_epsilon_node)

        # Calculate sqrt
        sqrt_name = s.get_unique_name("sqrt")
        sqrt_node = helper.make_node(
            "Sqrt",
            inputs=[add_epsilon_name],
            outputs=[sqrt_name],
            name=s.get_unique_name("sqrt"),
        )
        s.add_node(sqrt_node)

        # Divide input by sqrt
        div_name = s.get_unique_name("div")
        div_node = helper.make_node(
            "Div",
            inputs=[input_name, sqrt_name],
            outputs=[div_name],
            name=s.get_unique_name("div"),
        )
        s.add_node(div_node)

        # Multiply by scale
        s.get_unique_name("mul")
        mul_node = helper.make_node(
            "Mul",
            inputs=[div_name, scale_name],
            outputs=[final_output_name],
            name=s.get_unique_name("mul"),
        )
        s.add_node(mul_node)

    @staticmethod
    def _rms_norm(x, scale, epsilon):
        nnx.rms_norm_p.multiple_results = False
        return nnx.rms_norm_p.bind(x, scale, epsilon=epsilon)

    @staticmethod
    def rms_norm(x, scale, epsilon):
        """Binding function for rms_norm."""
        return RMSNormPlugin._rms_norm(x, scale, epsilon)

    @staticmethod
    def get_monkey_patch():
        """Returns a patched version of RMSNorm.__call__."""

        def patched_rms_norm_call(self, x):
            return RMSNormPlugin._rms_norm(
                x,
                self.scale.value,
                epsilon=self.epsilon,
            )

        return patched_rms_norm_call

    @staticmethod
    def patch_info():
        """Provides patching information."""
        return {
            "patch_targets": [nnx.RMSNorm],
            "patch_function": lambda _: RMSNormPlugin.get_monkey_patch(),
            "target_attribute": "__call__",
        }


# Register abstract evaluation function.
nnx.rms_norm_p.def_abstract_eval(RMSNormPlugin.abstract_eval)



import jax
import numpy as np
from typing import TYPE_CHECKING, Tuple, Sequence
from onnx import helper
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter


def _compute_conv_output_shape(
    x_shape: Tuple[int, ...],
    kernel_shape: Tuple[int, ...],
    strides: Sequence[int] | int,
    padding: str,
) -> Tuple[int, ...]:
    if isinstance(strides, int):
        strides = (strides, strides)
    N, H, W, _ = x_shape
    filter_height, filter_width, _, out_channels = kernel_shape
    if padding.upper() == "VALID":
        out_H = (H - filter_height) // strides[0] + 1
        out_W = (W - filter_width) // strides[1] + 1
    elif padding.upper() == "SAME":
        out_H = -(-H // strides[0])
        out_W = -(-W // strides[1])
    else:
        raise ValueError("Unsupported padding: " + padding)
    return (N, out_H, out_W, out_channels)


def compute_same_pads(input_size, filter_size, stride):
    out_size = int(np.ceil(float(input_size) / float(stride)))
    pad_total = max((out_size - 1) * stride + filter_size - input_size, 0)
    pad_before = pad_total // 2
    pad_after = pad_total - pad_before
    return pad_before, pad_after


@register_primitive(
    jaxpr_primitive=jax.lax.conv_general_dilated_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.conv.html",
    onnx=[
        {
            "component": "Conv",
            "doc": "https://onnx.ai/onnx/operators/onnx__Conv.html",
        }
    ],
    since="v0.2.0",
    context="primitives.lax",
    testcases=[
        {
            "testcase": "conv",  # NCHW & OIHW: no transposition needed.
            "callable": lambda x, y: jax.lax.conv(
                x, y, window_strides=(1, 1), padding="VALID"
            ),
            "input_shapes": [(1, 2, 3, 3), (1, 2, 2, 2)],
        },
        {
            "testcase": "conv2",  # NHWC & HWIO: transposition required.
            "callable": lambda x, y: jax.lax.conv_general_dilated(
                x,
                y,
                window_strides=(1, 1),
                padding="VALID",
                dimension_numbers=("NHWC", "HWIO", "NHWC"),
            ),
            "input_shapes": [(1, 3, 3, 2), (2, 2, 2, 1)],
        },
    ],
)
class ConvGeneralDilatedPlugin(PrimitivePlugin):
    """
    Plugin for converting jax.lax.conv_general_dilated to ONNX.
    """

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX conv_general_dilated primitive."""
        input_name = s.get_name(node_inputs[0])
        filter_var = node_inputs[1]
        output_name = s.get_name(node_outputs[0])

        dimension_numbers = params["dimension_numbers"]
        window_strides = params["window_strides"]
        padding = params["padding"]

        # Handle dimension numbers.
        lhs_spec, rhs_spec, out_spec = dimension_numbers
        if lhs_spec == (0, 3, 1, 2) and rhs_spec == (3, 2, 0, 1):  # NHWC & HWIO
            input_perm = [0, 3, 1, 2]  # NHWC -> NCHW
            kernel_perm = [3, 2, 0, 1]  # HWIO -> OIHW
            output_perm = [0, 2, 3, 1]  # NCHW -> NHWC
        elif lhs_spec == (0, 1, 2, 3) and rhs_spec == (0, 1, 2, 3):  # NCHW & OIHW
            input_perm = None
            kernel_perm = [0, 1, 2, 3]  # already OIHW
            output_perm = None  # already NCHW
        else:
            raise ValueError(f"Unhandled dimension_numbers: {dimension_numbers}")

        # Transpose input if necessary.
        conv_input = s.get_name(node_inputs[0])
        if input_perm:
            transposed_input = s.get_unique_name("input_transposed")
            s.add_node(
                helper.make_node(
                    "Transpose",
                    inputs=[input_name],
                    outputs=[transposed_input],
                    perm=input_perm,
                    name=s.get_unique_name("Transpose_input"),
                )
            )
            conv_input = transposed_input

        # Handle kernel: constant branch or dynamic branch.
        filter_name = s.get_name(filter_var)
        if filter_name in s.name_to_const:
            kernel_const = s.name_to_const[filter_name]
            kernel_transposed = np.transpose(kernel_const, kernel_perm)
            transposed_kernel_name = s.get_constant_name(kernel_transposed)
            s.name_to_const[transposed_kernel_name] = kernel_transposed
            kernel_shape = kernel_transposed.shape[
                2:
            ]  # OIHW: spatial dims at indices 2,3
        else:
            # Dynamic branch: insert a Transpose node for the kernel.
            transposed_kernel_name = s.get_unique_name("kernel_transposed")
            s.add_node(
                helper.make_node(
                    "Transpose",
                    inputs=[filter_name],
                    outputs=[transposed_kernel_name],
                    perm=kernel_perm,
                    name=s.get_unique_name("Transpose_kernel"),
                )
            )
            # For dynamic kernels: if rhs_spec is HWIO, spatial dims are the first two elements;
            # if rhs_spec is OIHW, they are the last two elements.
            if rhs_spec == (3, 2, 0, 1):
                kernel_shape = filter_var.aval.shape[
                    :2
                ]  # HWIO: (H, W, I, O) → spatial dims H, W
            else:
                kernel_shape = filter_var.aval.shape[2:]

        # Process padding.
        if isinstance(padding, str):
            if padding.upper() == "VALID":
                pads = [0, 0, 0, 0]
            elif padding.upper() == "SAME":
                # Compute SAME padding based on input and kernel shapes.
                # For simplicity, assume symmetric padding.
                if lhs_spec == (0, 3, 1, 2):  # Input in NHWC format
                    H_in, W_in = node_inputs[0].aval.shape[1:3]
                else:  # NCHW
                    H_in, W_in = node_inputs[0].aval.shape[2:4]
                filter_H, filter_W = kernel_shape
                pad_top, pad_bottom = compute_same_pads(
                    H_in, filter_H, window_strides[0]
                )
                pad_left, pad_right = compute_same_pads(
                    W_in, filter_W, window_strides[1]
                )
                pads = [pad_top, pad_left, pad_bottom, pad_right]
            else:
                raise ValueError("Unsupported padding string: " + padding)
        else:
            pads = [pad for pair in padding for pad in pair]

        conv_output = s.get_unique_name("conv_output")
        conv_node = helper.make_node(
            "Conv",
            inputs=[conv_input, transposed_kernel_name],
            outputs=[conv_output],
            kernel_shape=kernel_shape,
            strides=window_strides,
            pads=pads,
            name=s.get_unique_name("Conv"),
        )
        s.add_node(conv_node)

        if output_perm:
            s.add_node(
                helper.make_node(
                    "Transpose",
                    inputs=[conv_output],
                    outputs=[output_name],
                    perm=output_perm,
                    name=s.get_unique_name("Transpose_output"),
                )
            )
        else:
            if conv_output != output_name:
                s.add_node(
                    helper.make_node(
                        "Identity",
                        inputs=[conv_output],
                        outputs=[output_name],
                        name=s.get_unique_name("Identity_output"),
                    )
                )
        s.add_shape_info(output_name, node_outputs[0].aval.shape)



import numpy as np
from jax import core
from jax.extend.core import Primitive
from onnx import helper
from flax import nnx
from typing import TYPE_CHECKING, Tuple, Union
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the Linear primitive
nnx.linear_p = Primitive("nnx.linear")
nnx.linear_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=nnx.linear_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/linear.html",
    onnx=[
        {
            "component": "Gemm",
            "doc": "https://onnx.ai/onnx/operators/onnx__Gemm.html",
        },
        {
            "component": "Reshape",
            "doc": "https://onnx.ai/onnx/operators/onnx__Reshape.html",
        },
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "linear_2d",
            "callable": nnx.Linear(
                in_features=128,
                out_features=64,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [
                (32, 10, 128)
            ],  # Higher-rank input with batch dims (32,10)
        },
        {
            "testcase": "linear",
            "callable": nnx.Linear(
                in_features=128,
                out_features=64,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [("B", 128)],
        },
    ],
)
class LinearPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.Linear to ONNX.
    """

    @staticmethod
    def _shape_linear(
        x_shape: Tuple[Union[int, str], ...],
        kernel_shape: Tuple[int, ...],
        dimension_numbers=None,
    ) -> dict:
        """
        Calculates the input/output shapes for a standard linear layer.

        For nnx.Linear we assume:
          - x has shape (batch, in_features) (or higher-rank with batch dims preceding feature dim)
          - kernel has shape (in_features, out_features)

        If dimension_numbers is not provided, we assume the standard contraction:
          - Contract on the last dimension of x and the first dimension of kernel.
        """
        if len(x_shape) < 1:
            raise ValueError("Input must have at least one dimension")

        # Default: contract last dim of x with first dim of kernel.
        if dimension_numbers is None:
            lhs_contract = (len(x_shape) - 1,)  # last dimension of input
            rhs_contract = (0,)  # first dimension of kernel
            dimension_numbers = ((lhs_contract, rhs_contract), ((), ()))

        ((lhs_contract, rhs_contract), _) = dimension_numbers
        # Normalize negative indices.
        lhs_contract = (lhs_contract[0] % len(x_shape),)
        rhs_contract = (rhs_contract[0] % len(kernel_shape),)

        # The feature dimensions come from the contracted dimensions.
        x_feature_dims = lhs_contract
        # All other dimensions are considered batch dimensions.
        x_batch_dims = [d for d in range(len(x_shape)) if d not in lhs_contract]

        # Compute feature size.
        x_feature_sizes = [x_shape[d] for d in x_feature_dims]
        x_feature_size = (
            np.prod(x_feature_sizes).item()
            if all(isinstance(dim, int) for dim in x_feature_sizes)
            else x_shape[-1]
        )

        # Compute batch shape and size.
        x_batch_sizes = [x_shape[d] for d in x_batch_dims]
        dynamic_dim = None
        for dim in x_batch_sizes:
            if isinstance(dim, str):
                dynamic_dim = dim
                break
        x_batch_size = (
            dynamic_dim if dynamic_dim is not None else np.prod(x_batch_sizes).item()
        )

        # For a standard Linear, kernel shape is (in_features, out_features)
        out_features = kernel_shape[1]
        new_kernel_shape = kernel_shape  # No reshaping needed.

        # The GEMM will operate on a flattened input of shape (batch, in_features)
        input_gemm_shape = (x_batch_size, x_feature_size)
        output_gemm_shape = (x_batch_size, out_features)
        output_shape = tuple(x_batch_sizes) + (out_features,)

        return {
            "input": x_shape,
            "input_gemm": input_gemm_shape,
            "output_gemm": output_gemm_shape,
            "output": output_shape,
            "new_kernel": new_kernel_shape,
        }

    @staticmethod
    def _is_noop_reshape(original_shape, target_shape):
        """Return True if target_shape is equivalent to original_shape,
        allowing for a dynamic (-1) in the first dimension.
        """
        if len(original_shape) != len(target_shape):
            return False
        # Compare all dimensions except possibly the first.
        return original_shape[1:] == target_shape[1:]

    @staticmethod
    def abstract_eval(x, kernel, bias, dimension_numbers=None):
        """Abstract evaluation function for Linear."""
        shapes = LinearPlugin._shape_linear(x.shape, kernel.shape, dimension_numbers)
        return core.ShapedArray(shapes["output"], x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of Linear to ONNX format."""
        # node_inputs: [x, kernel, bias]
        x_var = node_inputs[0]
        kernel_var = node_inputs[1]
        bias_var = node_inputs[2]

        dimension_numbers = params.get("dimension_numbers")
        input_name = s.get_name(x_var)
        kernel_name = s.get_name(kernel_var)
        bias_name = s.get_name(bias_var)
        output_name = s.get_name(node_outputs[0])

        shapes = LinearPlugin._shape_linear(
            x_var.aval.shape, kernel_var.aval.shape, dimension_numbers
        )
        input_gemm_shape = shapes["input_gemm"]
        output_gemm_shape = shapes["output_gemm"]
        output_shape = shapes["output"]
        new_kernel_shape = shapes["new_kernel"]

        # Retrieve the kernel constant and reshape it if necessary.
        kernel_const = s.name_to_const[kernel_name]
        weights_name = s.builder.get_constant_name(
            kernel_const.reshape(new_kernel_shape)
        )

        # Determine target reshape shape for input.
        target_input_shape = tuple([-1] + list(input_gemm_shape[1:]))
        if LinearPlugin._is_noop_reshape(x_var.aval.shape, target_input_shape):
            input_reshape_name = input_name
        else:
            input_reshape_name = s.get_unique_name("input_reshape")
            reshape_shape_input = np.array(target_input_shape, dtype=np.int64)
            reshape_shape_input_name = s.builder.get_constant_name(reshape_shape_input)
            reshape_input_node = helper.make_node(
                "Reshape",
                inputs=[input_name, reshape_shape_input_name],
                outputs=[input_reshape_name],
                name=s.get_unique_name("reshape_input"),
                allowzero=0,
            )
            s.add_node(reshape_input_node)
            s.add_shape_info(input_reshape_name, input_gemm_shape)

        # Gemm node.
        gemm_output_name = s.get_unique_name("gemm_output")
        gemm_node = helper.make_node(
            "Gemm",
            inputs=[input_reshape_name, weights_name, bias_name],
            outputs=[gemm_output_name],
            name=s.get_unique_name("gemm"),
        )
        s.add_node(gemm_node)
        s.add_shape_info(gemm_output_name, output_gemm_shape)

        # Final reshape: restore to the original output shape if necessary.
        target_output_shape = [-1] + list(output_shape[1:])
        if LinearPlugin._is_noop_reshape(output_gemm_shape, tuple(target_output_shape)):
            # Instead of inserting an extra node, update the variable mapping and builder outputs.
            s.var_to_name[node_outputs[0]] = gemm_output_name
            # Update the corresponding output in the builder.
            for i, out in enumerate(s.builder.outputs):
                if out.name == output_name:
                    s.builder.outputs[i] = helper.make_tensor_value_info(
                        gemm_output_name,
                        out.type.tensor_type.elem_type,
                        [dim.dim_value for dim in out.type.tensor_type.shape.dim],
                    )
                    break
        else:
            reshape_output_shape = np.array(target_output_shape, dtype=np.int64)
            reshape_output_shape_name = s.builder.get_constant_name(
                reshape_output_shape
            )
            reshape_output_node = helper.make_node(
                "Reshape",
                inputs=[gemm_output_name, reshape_output_shape_name],
                outputs=[output_name],
                name=s.get_unique_name("reshape_output"),
                allowzero=0,
            )
            s.add_node(reshape_output_node)

    @staticmethod
    def _linear(x, kernel, bias, dimension_numbers=None):
        """Defines the primitive binding for Linear."""
        # Set default dimension numbers for a standard linear layer.
        if dimension_numbers is None:
            lhs_contract = (-1,)  # last dim of x
            rhs_contract = (0,)  # first dim of kernel
            dimension_numbers = ((lhs_contract, rhs_contract), ((), ()))

        return nnx.linear_p.bind(x, kernel, bias, dimension_numbers=dimension_numbers)

    @staticmethod
    def get_monkey_patch():
        """Returns a patched version of Linear's call method."""

        def patched_linear_call(self, x):
            # For nnx.Linear the standard contraction is last dim of input with first dim of kernel.
            lhs_contract = (-1,)
            rhs_contract = (0,)
            dimension_numbers = ((lhs_contract, rhs_contract), ((), ()))
            return LinearPlugin._linear(
                x,
                self.kernel.value,
                self.bias.value,
                dimension_numbers=dimension_numbers,
            )

        return patched_linear_call

    @staticmethod
    def patch_info():
        """Provides patching information for Linear."""
        return {
            "patch_targets": [nnx.Linear],
            "patch_function": lambda _: LinearPlugin.get_monkey_patch(),
            "target_attribute": "__call__",
        }


# Register abstract evaluation function.
nnx.linear_p.def_abstract_eval(LinearPlugin.abstract_eval)



import numpy as np
from jax import core
from jax.extend.core import Primitive
from onnx import helper
from flax import nnx
from typing import TYPE_CHECKING
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the JAX primitive for dot product attention.
nnx.dot_product_attention_p = Primitive("nnx.dot_product_attention")
nnx.dot_product_attention_p.multiple_results = False  # Correctly set at initialization


@register_primitive(
    jaxpr_primitive=nnx.dot_product_attention_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/attention.html#flax.nnx.dot_product_attention",
    onnx=[
        {
            "component": "Constant",
            "doc": "https://onnx.ai/onnx/operators/onnx__Constant.html",
        },
        {
            "component": "Einsum",
            "doc": "https://onnx.ai/onnx/operators/onnx__Einsum.html",
        },
        {
            "component": "Mul",
            "doc": "https://onnx.ai/onnx/operators/onnx__Mul.html",
        },
        {
            "component": "Softmax",
            "doc": "https://onnx.ai/onnx/operators/onnx__Softmax.html",
        },
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "dot_product_attention",
            "callable": lambda q, k, v: nnx.dot_product_attention(q, k, v),
            "input_shapes": [(2, 4, 8, 32), (2, 4, 8, 32), (2, 4, 8, 32)],
        },
    ],
)
class DotProductAttentionPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.dot_product_attention to ONNX.
    """

    @staticmethod
    def _shape_dot_product_attention(q_shape, k_shape, v_shape):
        """Computes the output shape of dot_product_attention."""
        return q_shape

    @staticmethod
    def abstract_eval(q, k, v, axis=-1):
        """Abstract evaluation function for dot_product_attention."""
        output_shape = DotProductAttentionPlugin._shape_dot_product_attention(
            q.shape, k.shape, v.shape
        )
        return core.ShapedArray(output_shape, q.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of dot_product_attention to ONNX format."""
        q_var, k_var, v_var = node_inputs[:3]
        output_var = node_outputs[0]

        q_name = s.get_name(q_var)
        k_name = s.get_name(k_var)
        v_name = s.get_name(v_var)
        output_name = s.get_name(output_var)

        q_shape = q_var.aval.shape
        k_shape = k_var.aval.shape
        v_shape = v_var.aval.shape
        B, N, H, E = q_shape
        _, M, _, _ = k_shape

        scale_value = 1.0 / np.sqrt(E)
        scale_const_name = s.builder.get_constant_name(
            np.array(scale_value, dtype=np.float32)
        )

        # Attention scores (Einsum)
        attn_scores_name = s.get_unique_name("attn_scores")
        attn_scores_shape = (B, N, H, M)
        s.add_node(
            helper.make_node(
                "Einsum",
                inputs=[q_name, k_name],
                outputs=[attn_scores_name],
                equation="BNHE,BMHE->BNHM",
                name=s.get_unique_name("einsum_qk"),
            )
        )
        s.add_shape_info(attn_scores_name, attn_scores_shape)

        # Scaled attention scores (Mul)
        scaled_scores_name = s.get_unique_name("scaled_scores")
        s.add_node(
            helper.make_node(
                "Mul",
                inputs=[attn_scores_name, scale_const_name],
                outputs=[scaled_scores_name],
                name=s.get_unique_name("scale"),
            )
        )
        s.add_shape_info(scaled_scores_name, attn_scores_shape)

        # Attention weights (Softmax)
        attn_weights_name = s.get_unique_name("attn_weights")
        s.add_node(
            helper.make_node(
                "Softmax",
                inputs=[scaled_scores_name],
                outputs=[attn_weights_name],
                axis=params.get("axis", -1),  # Correctly use 'axis' from params
                name=s.get_unique_name("softmax"),
            )
        )
        s.add_shape_info(attn_weights_name, attn_scores_shape)

        # Attention output (Einsum)
        attn_output_shape = (B, N, H, v_shape[-1])
        s.add_node(
            helper.make_node(
                "Einsum",
                inputs=[attn_weights_name, v_name],
                outputs=[output_name],
                equation="BNHM,BMHE->BNHE",
                name=s.get_unique_name("einsum_weights_v"),
            )
        )
        s.add_shape_info(output_name, attn_output_shape)

    @staticmethod
    def _dot_product_attention(q, k, v, axis=-1):
        """Defines the primitive binding for dot_product_attention."""
        return nnx.dot_product_attention_p.bind(q, k, v, axis=axis)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for dot_product_attention."""

        def patched_dot_product_attention(q, k, v, axis=-1):
            return DotProductAttentionPlugin._dot_product_attention(q, k, v, axis)

        return patched_dot_product_attention

    @staticmethod
    def patch_info():
        """Provides patching information for dot_product_attention."""
        return {
            "patch_targets": [nnx],
            "patch_function": lambda _: DotProductAttentionPlugin.get_monkey_patch(),
            "target_attribute": "dot_product_attention",
        }


# Register abstract evaluation function.
nnx.dot_product_attention_p.def_abstract_eval(DotProductAttentionPlugin.abstract_eval)



from jax import core, numpy as jnp
from jax.extend.core import Primitive
from onnx import helper
from typing import TYPE_CHECKING, Tuple, Union, Optional
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

import numpy as np

# Define a new primitive for squeeze
jnp.squeeze_p = Primitive("jnp.squeeze")
jnp.squeeze_p.multiple_results = False


@register_primitive(
    jaxpr_primitive=jnp.squeeze_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.squeeze.html",
    onnx=[
        {
            "component": "Squeeze",
            "doc": "https://onnx.ai/onnx/operators/onnx__Squeeze.html",
        }
    ],
    since="v0.1.0",
    context="primitives.jnp",
    testcases=[
        {
            "testcase": "squeeze_single_dim",
            "callable": lambda a: jnp.squeeze(a, axis=0),
            "input_shapes": [(1, 49, 10)],
        },
        {
            "testcase": "squeeze_multiple_dims",
            "callable": lambda a: jnp.squeeze(a, axis=(0, 2)),
            "input_shapes": [(1, 49, 1, 10)],
        },
        {
            "testcase": "squeeze_vit_output",
            "callable": lambda a: jnp.squeeze(a, axis=1),
            "input_shapes": [(1, 1, 10)],
        },
        {
            "testcase": "squeeze_dynamic_batch",
            "callable": lambda a: jnp.squeeze(a, axis=1),
            "input_shapes": [("B", 1, 10)],
        },
        {
            "testcase": "squeeze_all_dims",
            "callable": lambda a: jnp.squeeze(a),
            "input_shapes": [(1, 1, 1)],
        },
        {
            "testcase": "squeeze_negative_axis",
            "callable": lambda a: jnp.squeeze(a, axis=-1),
            "input_shapes": [(1, 49, 1)],
        },
        {
            "testcase": "squeeze_negative_axis_tuple",
            "callable": lambda a: jnp.squeeze(a, axis=(-1, -3)),
            "input_shapes": [(1, 49, 1)],
        },
        {
            "testcase": "squeeze_dynamic_and_negative_axis",
            "callable": lambda a: jnp.squeeze(a, axis=(-1, -3)),
            "input_shapes": [(1, "B", 1)],
        },
    ],
)
class SqueezePlugin(PrimitivePlugin):
    """
    Plugin for converting jax.numpy.squeeze to ONNX.
    """

    @staticmethod
    def _squeeze_abstract_eval(x, axes: Optional[Tuple[int, ...]]):
        """
        Compute the output shape for squeeze.
        - If no axes are provided, squeeze all dimensions that are 1.
        - If axes are provided, only remove dimensions that are concretely 1.
        - Dynamic dimensions (strings) are not squeezed.
        """
        x_shape = list(x.shape)
        if axes is None:
            new_shape = tuple(
                dim for dim in x_shape if not (isinstance(dim, int) and dim == 1)
            )
        else:
            normalized_axes = [
                axis if axis >= 0 else axis + len(x_shape) for axis in axes
            ]
            for axis in normalized_axes:
                if axis >= len(x_shape):
                    raise ValueError(f"Invalid axis {axis} for shape {x_shape}")
                if isinstance(x_shape[axis], int) and x_shape[axis] != 1:
                    raise ValueError(
                        f"Cannot squeeze dimension {axis} of shape {x_shape}: size is not 1."
                    )
            new_shape = tuple(
                dim for i, dim in enumerate(x_shape) if i not in normalized_axes
            )
        return core.ShapedArray(new_shape, x.dtype)

    @staticmethod
    def abstract_eval(x, axes: Optional[Tuple[int, ...]]):
        return SqueezePlugin._squeeze_abstract_eval(x, axes)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles ONNX conversion for jnp.squeeze."""
        axes = params["axes"]
        input_name = s.get_name(node_inputs[0])
        output_name = s.get_name(node_outputs[0])
        input_shape = node_inputs[0].aval.shape

        valid_axes = [axis if axis >= 0 else axis + len(input_shape) for axis in axes]

        # Create an initializer for axes (ONNX expects these as a tensor)
        axes_name = s.get_unique_name("squeeze_axes")
        s.add_initializer(name=axes_name, vals=np.array(valid_axes, dtype=np.int64))

        squeeze_node = helper.make_node(
            "Squeeze",
            inputs=[input_name, axes_name],
            outputs=[output_name],
            name=s.get_unique_name("squeeze"),
        )
        s.add_node(squeeze_node)

        output_shape = tuple(
            dim
            for i, dim in enumerate(input_shape)
            if i not in valid_axes
            or (isinstance(dim, str))  # Keep dynamic dims even if in valid_axes
        )

        s.add_shape_info(output_name, output_shape)

    @staticmethod
    def _squeeze(a, axis: Optional[Union[int, Tuple[int, ...]]] = None):
        """Defines the primitive binding for Squeeze."""
        if axis is None:
            axes = tuple(
                i for i, dim in enumerate(a.shape) if isinstance(dim, int) and dim == 1
            )
        elif isinstance(axis, int):
            axes = (axis,)
        else:
            axes = tuple(axis)
        return jnp.squeeze_p.bind(a, axes=axes)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for Squeeze."""

        def patched_squeeze(a, axis: Optional[Union[int, Tuple[int, ...]]] = None):
            return SqueezePlugin._squeeze(a, axis)

        return patched_squeeze

    @staticmethod
    def patch_info():
        """Provides patching information for Squeeze."""
        return {
            "patch_targets": [jnp],
            "patch_function": lambda _: SqueezePlugin.get_monkey_patch(),
            "target_attribute": "squeeze",
        }


# Register abstract evaluation function
jnp.squeeze_p.def_abstract_eval(SqueezePlugin.abstract_eval)



from jax import core
from jax.extend.core import Primitive
from flax import nnx
from onnx import helper
from typing import TYPE_CHECKING
from jax2onnx.plugin_system import register_primitive, PrimitivePlugin

if TYPE_CHECKING:
    from jax2onnx.converter.converter import Jaxpr2OnnxConverter

# Define the LeakyReLU primitive
nnx.leaky_relu_p = Primitive("nnx.leaky_relu")
nnx.leaky_relu_p.multiple_results = False  # Correctly set at initialization


@register_primitive(
    jaxpr_primitive=nnx.leaky_relu_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.nn.leaky_relu.html",
    onnx=[
        {
            "component": "LeakyRelu",
            "doc": "https://onnx.ai/onnx/operators/onnx__LeakyRelu.html",
        }
    ],
    since="v0.1.0",
    context="primitives.nnx",
    testcases=[
        {
            "testcase": "leaky_relu",
            "callable": lambda x: nnx.leaky_relu(x),
            "input_shapes": [(3,)],
        }
    ],
)
class LeakyReluPlugin(PrimitivePlugin):
    """
    Plugin for converting flax.nnx.leaky_relu to ONNX.
    """

    @staticmethod
    def abstract_eval(x, negative_slope=0.01):
        """Abstract evaluation function for LeakyReLU."""
        return core.ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of LeakyReLU to ONNX format."""
        input_var = node_inputs[0]
        output_var = node_outputs[0]

        input_name = s.get_name(input_var)
        output_name = s.get_name(output_var)

        # Retrieve the negative_slope parameter (defaulting to 0.01 if not provided)
        negative_slope = params.get("negative_slope", 0.01)

        leaky_relu_node = helper.make_node(
            "LeakyRelu",
            inputs=[input_name],
            outputs=[output_name],
            name=s.get_unique_name("leaky_relu"),
            alpha=negative_slope,
        )
        s.add_node(leaky_relu_node)

    @staticmethod
    def _leaky_relu(x, negative_slope=0.01):
        """Defines the primitive binding for LeakyReLU."""
        return nnx.leaky_relu_p.bind(x, negative_slope=negative_slope)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for LeakyReLU."""

        def patched_leaky_relu(x, negative_slope=0.01):
            return LeakyReluPlugin._leaky_relu(x, negative_slope)

        return patched_leaky_relu

    @staticmethod
    def patch_info():
        """Provides patching information for LeakyReLU."""
        return {
            "patch_targets": [nnx],
            "patch_function": lambda _: LeakyReluPlugin.get_monkey_patch(),
            "target_attribute": "leaky_relu",
        }


# Register abstract evaluation function
nnx.leaky_relu_p.def_abstract_eval(LeakyReluPlugin.abstract_eval)
